"""
Calculating two-point functions, like MSD & ACF

Note that all of these (MSD, ACF, VACF, ...) contain the same information. See
also the python package ``bayesmsd``.
"""

import warnings

import numpy as np

from ..trajectory import Trajectory
from ..taggedset import TaggedSet

DEFKEY='P2'

def P2traj(traj, TA=True, recalculate=False,
           function=None, preproc=None, postproc=None,
           writeto=DEFKEY,
           ):
    """
    Calculate a two-point function for a single trajectory

    The results are stored in the ``traj.meta[writeto]``. By default, if this
    field exists already, the function aborts. To recalculate and overwrite,
    set ``recalculate = True``.

    Parameters
    ----------
    traj : Trajectory
        the trajectory for which to calculate the MSD
    TA : bool, optional
        whether to time average

    Other Parameters
    ----------------
    recalculate : bool
        set to ``True`` to ensure that the calculation is actually performed
    function : {'SD', 'D', 'SP'} or callable
        the function to evaluate. Should be ``fun(traj[m], traj[n]) -->
        float``, where ``m >= n``, and it should be vectorized (i.e. work on
        numpy arrays and return the corresponding arrays).
    preproc : callable or None
        will be applied to the trajectory before processing. Example: ``lambda
        traj: traj.diff()`` gives the increment trajectory
    postproc : callable or None
        will be applied to the final result, i.e. should take a (T,) array and
        return such. Relevant when accessing this function through
        ``P2dataset``, since in that case this `!postproc` is applied _before_
        averaging. Use case: normalizing correlation functions
    writeto : hashable or None
        where to store the output of the calculation in the ``traj.meta`` dict.
        Set to ``None`` to return the output dict instead of storing it.
        Defaults to ``'P2'``.

    Returns
    -------
    dict, optional
        A dict with keys ``'data', 'N'`` giving the averaged data and the count
        of valid data points for each lag time. This is usually written into
        the ``traj.meta`` dict, but returned if ``writeto is None``.

    See also
    --------
    P2dataset, MSD

    Notes
    -----
    Explicitly, the ``recalculate`` parameter is equivalent to
    >>> del traj.meta[writeto]
    """
    if recalculate:
        for key in [writeto]:
            try:
                del traj.meta[key]
            except:
                pass
    
    if writeto not in traj.meta.keys():

        if callable(preproc):
            proc_traj = preproc(traj)
        else:
            proc_traj = traj

        if TA:
            data = [function(proc_traj[:], proc_traj[:])]
            data += [function(proc_traj[i:], proc_traj[:-i]) for i in range(1, len(proc_traj))]
        else:
            istart = np.min(np.nonzero(~np.any(np.isnan(proc_traj.data), (0, 2)))[0])
            data = [[function(proc_traj[i], proc_traj[istart])] for i in range(istart, len(proc_traj))]

        del proc_traj # just for clarity

        with warnings.catch_warnings():
            warnings.filterwarnings(action='ignore', message='Mean of empty slice')

            out = {
                'N' : np.array([np.count_nonzero(~np.isnan(dat)) for dat in data]),
                'data' : np.array([np.nanmean(dat) for dat in data]),
            }
            if callable(postproc):
                out['data'] = postproc(out['data'])

        if writeto is None:
            return out
        else:
            traj.meta[writeto] = out

def P2dataset(dataset, givevar=False, giveN=False, average_in_logspace=False, **kwargs):
    """
    Ensemble average two-point functions

    Parameters
    ----------
    dataset : TaggedSet
        a list of `Trajectory`

    Other Parameters
    ----------------
    givevar : bool, optional
        whether to also return the variance around the mean
    giveN : bool, optional
        whether to return the sample size for each MSD data point
    average_in_logspace : bool, optional
        set to ``True`` to replace the arithmetic with a geometric mean.
    kwargs : keyword arguments
        are all forwarded to forwarded to `P2traj`, see that docstring.

    Returns
    -------
    eP2 : np.ndarray
        the calculated ensemble mean
    var : np.ndarray, optional
        variance around the mean
    N : np.ndarray, optional
        number of data points going into each estimate

    See also
    --------
    P2traj, MSD, ACF
    """
    try:
        p2key = kwargs['writeto']
    except KeyError:
        p2key = DEFKEY

    for traj in dataset:
        P2traj(traj, **kwargs)

    P2s = [traj.meta[p2key]['data'] for traj in dataset]
    Ns = [traj.meta[p2key]['N'] for traj in dataset]

    maxlen = max(len(P2) for P2 in P2s)
    allP2 = np.empty((len(P2s), maxlen), dtype=float)
    allP2[:] = np.nan
    allN = np.zeros((len(Ns), maxlen), dtype=int)
    for i, (P2, N) in enumerate(zip(P2s, Ns)):
        allP2[i, :len(P2)] = P2
        allN[i, :len(N)] = N
    allN[np.where(np.isnan(allP2))] = 0

    if average_in_logspace: # pragma: no cover
        allP2 = np.log(allP2[:, 1:])
        N0 = np.sum(allN[:, 0])
        allN = allN[:, 1:]

    N = np.sum(allN, axis=0)
    with warnings.catch_warnings():
        warnings.filterwarnings(action='ignore', message=r'(invalid value|divide by zero) encountered in true_divide')
        meanN = N / np.sum(allN != 0, axis=0)
        eP2 = np.nansum(allP2*allN, axis=0) / N
        if givevar:
            var = np.nansum((allP2-eP2)**2 * allN, axis=0) / (N-meanN)

    if average_in_logspace: # pragma: no cover
        eP2 = np.insert(np.exp(eP2), 0, 0)
        if givevar:
            var = np.insert(np.exp(var), 0, 0)
        N = np.insert(N, 0, N0)

    if givevar and giveN: # pragma: no cover
        return eP2, var, N
    elif givevar:
        return eP2, var
    elif giveN:
        return eP2, N
    else:
        return eP2

def P2(*args, **kwargs):
    """
    Shortcut function to calculate P2s.

    Will select either `P2traj` or `P2dataset`, depending on the type of the
    first argument. Everything is then forwarded to that function.

    This is mostly just a template for the following functions MSD, ACF, ...

    See also
    --------
    P2traj, P2dataset
    """
    if issubclass(type(args[0]), Trajectory):
        P2traj(*args, **kwargs)
        if 'writeto' in kwargs:
            writeto = kwargs['writeto']
        else:
            writeto = DEFKEY
        return args[0].meta[writeto]['data']
    elif issubclass(type(args[0]), TaggedSet):
        return P2dataset(*args, **kwargs)
    else: # pragma: no cover
        raise ValueError("Did not understand first argument, with type {}".format(type(args[0])))

def MSD(*args, **kwargs):
    """
    Calculate MSD for a `Trajectory` or dataset.

    This is an implementation of `P2`. See `P2traj` and `P2dataset` for
    detailed description of parameters, here we will just list the most
    interesting / relevant ones

    Parameters
    ----------
    in : Trajectory or TaggedSet of Trajectory
        the first argument is the input data, either a single `Trajectory`, or
        a dataset that will be ensemble averaged over
    TA : bool, optional
        whether to time average on (each) single trajectory
    recalculate : bool, optional
        set to ``True`` to ensure that the calculation is actually performed.
        Otherwise, this function might just return precalculated values from
        ``traj.meta['MSD']``.
    giveN : bool, optional
        only applicable when `!in` is a dataset. Whether to also return the
        sample size for each MSD data point.

    Returns
    -------
    msd : np.ndarray
        the calculated MSD. Note that ``msd[i] = MSD(iΔt)``, i.e. ``msd[0] =
        0``.
    N : np.ndarray, optional
        the sample size for each point, see `!giveN`.

    See also
    --------
    P2traj, P2dataset
    """
    def SD(xm, xn):
        return np.sum((xm-xn)**2, axis=-1)

    return P2(*args, **kwargs, function=SD, writeto='MSD')

def ACov(*args, **kwargs):
    """
    Calculate autocovariance for a `Trajectory` or dataset.

    This is an implementation of `P2`. See `P2traj` and `P2dataset` for
    detailed description of parameters, here we will just list the most
    interesting / relevant ones

    Parameters
    ----------
    in : Trajectory or TaggedSet of Trajectory
        the first argument is the input data, either a single `Trajectory`, or
        a dataset that will be ensemble averaged over
    TA : bool, optional
        whether to time average on (each) single trajectory
    recalculate : bool, optional
        set to ``True`` to ensure that the calculation is actually performed.
        Otherwise, this function might just return precalculated values from
        ``traj.meta['ACov']``.
    giveN : bool, optional
        only applicable when `!in` is a dataset. Whether to also return the
        sample size for each data point.

    Returns
    -------
    acov : np.ndarray
        the calculated autocovariance
    N : np.ndarray, optional
        the sample size for each point, see `!giveN`

    See also
    --------
    MSD, P2traj, P2dataset
    """
    def SP(xm, xn):
        return np.sum(xm*xn, axis=-1)

    return P2(*args, **kwargs, function=SP, writeto='ACov')

def ACorr(*args, **kwargs):
    """
    Calculate autocorrelation for a `Trajectory` or dataset.

    This is an implementation of `P2`. See `P2traj` and `P2dataset` for
    detailed description of parameters, here we will just list the most
    interesting / relevant ones

    Parameters
    ----------
    in : Trajectory or TaggedSet of Trajectory
        the first argument is the input data, either a single `Trajectory`, or
        a dataset that will be ensemble averaged over
    TA : bool, optional
        whether to time average on (each) single trajectory
    recalculate : bool, optional
        set to ``True`` to ensure that the calculation is actually performed.
        Otherwise, this function might just return precalculated values from
        ``traj.meta['ACorr']``.
    giveN : bool, optional
        only applicable when `!in` is a dataset. Whether to also return the
        sample size for each data point.

    Returns
    -------
    acorr : np.ndarray
        the calculated autocorrelation
    N : np.ndarray, optional
        the sample size for each point, see `!giveN`

    See also
    --------
    MSD, P2traj, P2dataset
    """
    def SP(xm, xn):
        return np.sum(xm*xn, axis=-1)
    def normalize(data):
        return data / data[0]

    return P2(*args, **kwargs, function=SP, postproc=normalize, writeto='ACorr')

def VACov(*args, **kwargs):
    """
    Calculate velocity autocovariance for a `Trajectory` or dataset.

    This is an implementation of `P2`. See `P2traj` and `P2dataset` for
    detailed description of parameters, here we will just list the most
    interesting / relevant ones

    Parameters
    ----------
    in : Trajectory or TaggedSet of Trajectory
        the first argument is the input data, either a single `Trajectory`, or
        a dataset that will be ensemble averaged over
    TA : bool, optional
        whether to time average on (each) single trajectory
    recalculate : bool, optional
        set to ``True`` to ensure that the calculation is actually performed.
        Otherwise, this function might just return precalculated values from
        ``traj.meta['VACov']``.
    giveN : bool, optional
        only applicable when `!in` is a dataset. Whether to also return the
        sample size for each data point.

    Returns
    -------
    vacov : np.ndarray
        the calculated velocity autocovariance
    N : np.ndarray, optional
        the sample size for each point, see `!giveN`

    See also
    --------
    MSD, P2traj, P2dataset
    """
    def SP(xm, xn):
        return np.sum(xm*xn, axis=-1)

    return P2(*args, **kwargs, function=SP, preproc=lambda traj: traj.diff(), writeto='VACov')

def VACorr(*args, **kwargs):
    """
    Calculate velocity autocorrelation for a `Trajectory` or dataset.

    This is an implementation of `P2`. See `P2traj` and `P2dataset` for
    detailed description of parameters, here we will just list the most
    interesting / relevant ones

    Parameters
    ----------
    in : Trajectory or TaggedSet of Trajectory
        the first argument is the input data, either a single `Trajectory`, or
        a dataset that will be ensemble averaged over
    TA : bool, optional
        whether to time average on (each) single trajectory
    recalculate : bool, optional
        set to ``True`` to ensure that the calculation is actually performed.
        Otherwise, this function might just return precalculated values from
        ``traj.meta['VACorr']``.
    giveN : bool, optional
        only applicable when `!in` is a dataset. Whether to also return the
        sample size for each data point.

    Returns
    -------
    vacorr : np.ndarray
        the calculated velocity autocorrelation
    N : np.ndarray, optional
        the sample size for each point, see `!giveN`

    See also
    --------
    MSD, P2traj, P2dataset
    """
    def SP(xm, xn):
        return np.sum(xm*xn, axis=-1)
    def normalize(data):
        return data / data[0]

    return P2(*args, **kwargs, function=SP, preproc=lambda traj: traj.diff(), postproc=normalize, writeto='VACorr')



"""
Loading data from common formats into the Trajectory and TaggedSet structures
used throughout the library
"""

import os,sys

import numpy as np
import h5py

from ..trajectory import Trajectory
from ..taggedset import TaggedSet
from . import hdf5 as hdf5_mod

def csv(filename, columns=['x', 'y', 't', 'id'], tags=None, meta_post={}, **kwargs):
    """
    Load data from a .csv file.

    This uses ``np.genfromtxt``, and all kwargs are forwarded to it. By
    default, we assume the delimiter ``','`` and utf8 encoding for string data,
    but these can of course be changed. Refer to ``numpy.genfromtxt``.
    
    Parameters
    ----------
    filename : string or file-like object
        the file to be read
    columns : list
        how to interpret the columns in the file. Use any of these identifiers:
        ``{'x', 'y', 'z', 'x2', 'y2', 'z2', 't', 'id', None}``, where ``'t'``
        (mandatory) is the frame number, ``'id'`` (mandatory) the trajectory
        id, and the remaining ones can be used to indicate spatial components
        of single or double-locus trajectories. Use ``None`` to indicate a
        column that should be ignored.  Columns beyond the list given here will
        be ignored in any case. Finally, the data for any str identifier not
        matching one of the above will be written to a corresponding entry in
        the trajectory's `meta` dict.
    tags : str, list of str or set of str, optional
        the tag(s) to be associated with trajectories from this file
    meta_post : dict, optional
        post-processing options for the `meta` data. Keys should be `meta`
        field names, values can be "unique" or "mean". With the former, all the
        values in the corresponding column should be the same, and only that
        value (instead of the whole array) will be written into the meta field.
        With the latter we simply take the mean of the array.

    Returns
    -------
    TaggedSet
        the loaded data set

    Examples
    --------
    This function can be used to load data from ``pandas.DataFrame`` tables, if
    they conform to the format described above:

    >>> import io
    ... import pandas as pd
    ... import noctiluca as nl
    ...
    ... # Set up a DataFrame containing some dummy data
    ... # Caveat to pay attention to: the order of the columns is important!
    ... df = pd.DataFrame()
    ... df['frame_no'] = [1, 2, 3]
    ... df['trajectory_id'] = [4, 4, 4]
    ... df['coord1'] = [1, 2, 3]
    ... df['coord2'] = [4, 5, 6]
    ...
    ... csv_stream = io.StringIO(df.to_csv())
    ... dataset = nl.io.load.csv(csv_stream,
    ...                          [None, 't', 'id', 'x', 'y'], # first column will be index
    ...                          delimiter=',',               # pandas' default
    ...                          skip_header=1,               # pandas prints a header line
    ...                         )
    """
    col_inds = {}
    for i, key in enumerate(columns):
        if type(key) == str: # make sure to exclude None's
            col_inds[key] = i

    keys = col_inds.keys()
    assert 'id' in keys
    assert 't' in keys
    
    # Get shape of trajectory and check that the given keys make sense
    if 'z' in keys:
        d = 3
        assert 'y' in keys
        assert 'x' in keys
    elif 'y' in keys:
        d = 2
        assert 'x' in keys
    elif 'x' in keys:
        d = 1
    else: # pragma: no cover
        raise ValueError("No valid coordinates found in specification: {}".format(columns))

    # data_keys = ['x', 'x2', 'y', 'y2', ...]
    data_keys = sorted(keys & {'x', 'y', 'z', 'x2', 'y2', 'z2'})
    N = 1
    if any('2' in key for key in keys):
        N = 2
        for key in data_keys:
            assert key[0] in data_keys
            assert key[0]+'2' in data_keys

    # ['id', 't', {data}, {meta}]
    # this is mostly to keep track of which keys exist
    sorted_keys = ['id', 't', data_keys,
                   keys - {'id', 't', *data_keys}]

    # Read data
    gft_kwargs = dict(delimiter=',', dtype=None, encoding='utf8')
    gft_kwargs.update(kwargs)
    data = np.genfromtxt(filename, **gft_kwargs)

    # This feels suboptimal... maybe there's a better way?
    data_cols = [col_inds[key] for key in sorted_keys[2]]
    meta_cols = [col_inds[key] for key in sorted_keys[3]]
    try:
        # sorted_data = [id-array, t-array, data-array, list of meta-arrays]
        sorted_data = [
            np.array([line[col_inds['id']] for line in data]),
            np.array([line[col_inds['t' ]] for line in data]).astype(int),
            np.array([[line[col] for col in data_cols]
                      for line in data
                     ]).astype(float), # shape: (-1, N*d), sorted x, x2, y, ...
            [np.array([line[col] for line in data]) for col in meta_cols],
        ]
    except IndexError:
        raise ValueError("Too many columns for file. Did you use the right delimiter?")
    del data
    ids = set(sorted_data[0])

    # Assemble data set
    out = TaggedSet()
    for myid in ids:
        ind = sorted_data[0] == myid
        mydata = np.moveaxis(sorted_data[2][ind].reshape((-1, d, N)), 2, 0)
        myt = sorted_data[1][ind]
        myt -= np.min(myt)

        trajdata = np.empty((N, np.max(myt)+1, d), dtype=float)
        trajdata[:] = np.nan
        trajdata[:, myt, :] = mydata

        meta = {}
        for i, key in enumerate(sorted_keys[3]):
            mymeta = sorted_data[3][i][ind]
            if key in meta_post:
                post = meta_post[key]
                if post == 'unique':
                    ms = set(mymeta)
                    if len(ms) > 1:
                        raise RuntimeError("Data in column '{}' is not unique for trajectory with id {}".format(key, myid))
                    meta[key] = ms.pop()
                elif post == 'mean':
                    meta[key] = np.mean(mymeta)
                elif post == 'nanmean':
                    meta[key] = np.nanmean(mymeta.astype(float))
                else: # pragma: no cover
                    raise ValueError(f"invalid meta post-proc: {post}")
            else: # assume that we have floats and fill with nan's
                meta[key] = np.empty(np.max(myt)+1, dtype=float)
                meta[key][:] = np.nan
                meta[key][myt] = mymeta

        out.add(Trajectory(trajdata, **meta), tags)

    return out

def evalSPT(filename, tags=set()):
    """
    Load data in the format used by evalSPT

    This is a shortcut for ``csv(filename, ['x', 'y', 't', 'id'], tags,
    delimiter='\t')``.

    See also
    --------
    csv
    """
    return csv(filename, ['x', 'y', 't', 'id'], tags, delimiter='\t')

def hdf5(filename, group='/'):
    """
    Load data from an HDF5 file

    Parameters
    ----------
    filename : str or pathlib.Path
        the file to read
    group : str
        which group in the file to read. Defaults to root, i.e. the whole file.

    Returns
    -------
    dict or object
        whatever is stored in the file
    """
    group, name = hdf5_mod.check_group_or_attr(group)
    with h5py.File(str(filename), 'r') as f:
        if name:
            return f[group].attrs[name]
        else:
            # Could still be an attribute and an unaware user
            try:
                return hdf5_mod.read(f[group])
            except KeyError:
                # Split by hand
                parts = group.split('/')
                group = '/'.join(parts[:-1])
                group = group if len(group) > 0 else '/'
                name = parts[-1]
                return f[group].attrs[name]



"""
Some functions for writing trajectories / data sets to file
"""
import numpy as np
import scipy.io
import csv as csv_mod
import h5py

from . import hdf5 as hdf5_mod

def csv(data, filename, header=True, delimiter='\t'):
    """
    A quick-and-dirty csv-writer. Might be updated eventually.

    Parameters
    ----------
    data : `TaggedSet` of `Trajectory`
        the data set to write to file
    filename : str
        the file to write to
    header : bool, optional
        whether to write a header line with column names
    delimiter : chr, optional
        which character to use as delimiter

    Notes
    -----
    The columns in the file will be ``['id', 'frame', 'x', 'y', 'z', 'x2',
    'y2', 'z2']``, where of course only those coordinates present in the data
    set will be written.

    Missing frames, i.e. those where all of the coordinates are ``np.nan`` will
    simply be omitted.

    Since `TaggedSet` and `Trajectory` have more structure than can reasonably
    represented in ``.csv`` files, this function has no aspirations of writing
    the whole structure to file. It can write only the "core" data, i.e. the
    actual trajectories.
    """
    N = data.map_unique(lambda traj : traj.N)
    d = data.map_unique(lambda traj : traj.d)

    with open(filename, 'w', newline='') as csvfile:
        writer = csv_mod.writer(csvfile, delimiter=delimiter)

        if header:
            head = ['id', 'frame']
            coords = ['x', 'y', 'z'][:d]
            if N == 2:
                coords += [c + '2' for c in coords]
            writer.writerow(head + coords)

        for traj_id, traj in enumerate(data):
            for frame in range(traj.T):
                if np.all(np.isnan(traj.data[:, frame, :])):
                    continue

                head = [traj_id, frame]
                coords = []
                for n in range(N):
                    coords += traj.data[n, frame, :].tolist()

                writer.writerow(head + coords)

def mat(data, filename):
    """
    Write a dataset to MATLAB's .mat format

    This will produce a cell array containing the individual trajectories as
    structs. All the meta-data is passed along as well. The tags associated
    with the trajectory will be written to an entry ``'noctiluca_tags'``.

    Parameters
    ----------
    data : TaggedSet of Trajectory
        the data set to write
    filename : str
        the file to write to
    """
    trajs = np.empty(len(data), dtype=object)
    for i, (traj, tags) in enumerate(data(giveTags=True)):
        traj.noctiluca_tags = list(tags)

        if traj.localization_error is None:
            traj.localization_error = np.nan
        if traj.parity is None:
            traj.parity = np.nan

        trajs[i] = traj

    scipy.io.savemat(filename, {'trajs' : trajs})

def hdf5(data, filename, group=None):
    """
    Write to HDF5 file

    Parameters
    ----------
    data : TaggedSet or dict
        the stuff to write
    filename : str or pathlib.Path
        where to write to
    group : str
        where in the file to write the data. If unspecified, the file will be
        truncated and content written to the root node.

    Notes
    -----
    Caution is advised, since this function will silently overwrite existing
    data (this is most often the desired behavior).
    """
    mode = 'a' # append to existing file or create new if nonexistent
    if group is None:
        mode = 'w' # overwrite file if exists
        group = '/'
        name = None
    else:
        if group.endswith('/'):
            group = group[:-1]

        parts = group.split('/')
        group = '/'.join(parts[:-1])
        name = parts[-1]

        group = group if len(group) > 0 else '/'
        assert len(name) > 0

    # After the above:
    #  - `group` will be a valid string
    #  - `name` might be None (only if group was None)

    with h5py.File(str(filename), mode) as f:
        try:
            f.create_group(group)
        except ValueError: # if group exists, specifically '/'
            pass

        # Silent overwrite
        if name is not None and name in f[group]:
            del f[group][name]

        hdf5_mod.write(data, name, f[group])

def hdf5_subTaggedSet(data, filename, group, refTaggedSet=None):
    """
    Write a subset of an already stored `!TaggedSet` to file

    Sometimes it is handy to store subsets of data in a directly loadable way
    (i.e. as its own `!TaggedSet` object). This would duplicate data and thus
    increase file size, so this function takes advantage of hdf5's hard links
    to store the properly pruned `!TaggedSet` by just linking to the
    corresponding entries in the full data set, which should be located in the
    same file at the `!refTaggedSet` address.

    Parameters
    ----------
    data : TaggedSet
        a `!TaggedSet` with some selection applied. The full data set
        (potentially with a different selection, this does not matter) should
        already be written to `!filename` under the path `!refTaggedSet`.
    filename : str or pathlib.Path
        the file to store things in
    group : str
        the location in the file where to store the new entry
    refTaggedSet : str
        where the full data set is stored in the file.

    Notes
    -----
    Caution is advised, since this function will silently overwrite existing
    data (this is most often the desired behavior).

    This function is intended for storing selections (subsets) of `!TaggedSets`
    such that they can be read from file as complete `!TaggedSets` themselves.
    Usecases include having a big dataset, out of which you routinely need only
    a specific part. If your subset is identified by the tag ``subset`` in the
    big dataset, this is equivalent to

    >>> from noctiluca import io
    ...
    ... big_data = io.load.hdf5('file_with_big_data.h5', 'data')
    ... big_data.makeSelection(tags='subset')
    ... data = big_data.copySelection()
    ...
    ... # This can be reduced to a single line by saving the selection beforehand
    ... # when saving the data:
    ... big_data.makeSelection()
    ... io.write.hdf5(big_data, 'file_with_big_data_and_subset.h5', 'data')
    ... big_data.makeSelection(tags='subset')
    ... io.write.hdf5_subTaggedSet(big_data, 'file_with_big_data_and_subset.h5',
    ...                            'data_subset', refTaggedSet='/data')
    ...
    ... # so now when loading the data, we can just do
    ... data = io.load.hdf5('file_with_big_data_and_subset.h5', 'data_subset')

    Note that basically we just shifted the process of making the selection
    from loading to writing. This however can come in very handy when the
    selection process is more involved than a simple tag, or you distribute
    your data to others, who will appreciate an easy way to load just the
    relevant data.
    """
    # Check arguments
    if type(refTaggedSet) != str:
        raise ValueError("Please specify where the full dataset is saved in the file")

    # Dissect group as group/name
    parts = group.split('/')
    group = '/'.join(parts[:-1])
    name = parts[-1]

    group = group if len(group) > 0 else '/'

    if len(name) == 0:
        raise ValueError("Please specify a name (group) for your new entry")

    # We rely on the internal structure of TaggedSet, so check that we're up to
    # date about that
    vars_and_types = {'_data' : list, '_tags' : list, '_selected' : list}
    assert vars(data).keys() == vars_and_types.keys()
    for var, typ in vars_and_types.items():
        assert type(getattr(data, var)) == typ

    # Prepare the data to be saved
    # _data is just assembled as hdf5 paths here, since these will be hard links
    pseudo_TaggedSet = dict()
    pseudo_TaggedSet['_data'] = [refTaggedSet+f'/_data/{i}' for i, sel in enumerate(data._selected) if sel]
    pseudo_TaggedSet['_tags'] = [t for t, sel in zip(data._tags, data._selected) if sel]
    pseudo_TaggedSet['_selected']= len(data)*[True]

    # Write
    with h5py.File(str(filename), 'a') as f:
        # First things first: make sure that the hard links work out
        refData = f[refTaggedSet+'/_data']
        if isinstance(refData, h5py.Dataset) or "0" not in refData:
            raise ValueError("Cannot save subset of TaggedSet of built-in type (these are not saved as groups, i.e. we cannot link to individual entries)")

        ls = pseudo_TaggedSet['_data']
        for i, path in enumerate(ls):
            ls[i] = f[path]

        # Set up new entry
        try:
            f.create_group(group)
        except ValueError:
            pass # group exists, that's okay

        # Silent overwrite
        if name in f[group]:
            del f[group][name]

        # Write everything to it
        hdf5_base = f[group].create_group(name)
        for key in f[refTaggedSet]:
            if key in pseudo_TaggedSet:
                hdf5_mod.write(pseudo_TaggedSet[key], key, hdf5_base)
            else: # pragma: no cover
                hdf5_base[key] = f[refTaggedSet][key] # hard link

        # Specifically, this copies _HDF5_ORIG_TYPE_
        for key, value in f[refTaggedSet].attrs.items():
            hdf5_base.attrs[key] = value



"""
Everything to do with cleaning up experimental data
"""

from copy import deepcopy
import numpy as np

from ..trajectory import Trajectory
from ..taggedset import TaggedSet

def split_trajectory_at_big_steps(traj, threshold):
    """
    Removes suspected mislinkages.

    Exception: two consecutive big steps such that the overall displacement is
    smaller than the threshold are single misconnections in otherwise fine
    trajectories. In this case we will simply remove the misconnected point.

    Parameters
    ----------
    traj : Trajectory
        the trajectory to investigate
    threshold : float
        the maximum allowed frame to frame displacement
    
    Returns
    -------
    set of Trajectory

    See also
    --------
    split_dataset_at_big_steps

    Notes
    -----
    As of now, this only works on trajectories with ``N=1``.

    This really just checks for frame-to-frame connections exceeding the
    threshold. So if there are missing frames in a trajectory, the step across
    those missing data will not be considered.
    """
    if traj.N != 1: # pragma: no cover
        raise ValueError("Cannot detect mislinkages in trajectories with N > 1")

    old_npwarns = np.seterr(invalid='ignore') # yes, we'll do np.nan <= threshold. Gives False.

    difftraj = traj.diff().abs()
    step_isBig = np.where(difftraj[:][:, 0] <= threshold, 0, 1.)
    step_isBig[np.where(np.isnan(difftraj[:][:, 0]))[0]] = np.nan

    np.seterr(**old_npwarns)
    
    step_isBig = np.pad(step_isBig, 1, constant_values=np.nan) # Now step_isBig[i] describes traj[i] - traj[i-1]
    inds_bigsteps = np.where(step_isBig == 1)[0]

    # Check for single misconnections
    for ind in inds_bigsteps:
        if ((step_isBig[(ind-1):(ind+2)] == 1).tolist() == [False, True, True]
            and step_isBig[ind+2] != 1):
            traj.data[:, ind, :] = np.nan
            step_isBig[ind:(ind+2)] = np.nan # Now the steps don't exist anymore
    
    # If now everything's fine, that's cool
    if not np.any(step_isBig == 1):
        return {traj}
    
    # Split at remaining big steps
    inds_bigsteps = np.where(step_isBig == 1)[0]
    new_trajs = set()
    old_ind = 0
    for ind in inds_bigsteps:
        new_trajs.add(Trajectory(traj.data[:, old_ind:ind, :]))
        old_ind = ind
    new_trajs.add(Trajectory(traj.data[:, old_ind:, :]))
    del traj
    
    return {traj for traj in new_trajs if len(traj) > 1}

def split_dataset_at_big_steps(data, threshold):
    """
    Apply `split_trajectory_at_big_steps` to a whole data set

    Parameters
    ----------
    data : TaggedSet
        the dataset
    threshold : float
        the maximum allowed step size

    Returns
    -------
    TaggedSet
        a new data set with the split trajectories

    See also
    --------
    split_trajectory_at_big_steps
    """
    def gen():
        for traj, tags in data(giveTags=True):
            for part_traj in split_trajectory_at_big_steps(traj, threshold):
                yield (deepcopy(part_traj), deepcopy(tags))
    
    return TaggedSet(gen())



import numpy as np
import re
import h5py

_TYPE_KW = '_HDF5_ORIG_TYPE_'

# The registry registers custom read/write functions
# It is a dict, where the key is a unique string identifier for the type to be
# handled and the entry is a tuple of
# 
#   (type, writer, reader)
# 
# where signatures for the writer and reader are
# 
#   hdf5_container = writer(object, name, hdf5_base_point)
#   object = reader(hdf5_container)
#
# Note that the signature of the writer function allows to pass methods from
# user-defined classes
#
# Writer functions should use the @hdf5.writer decorator (which writes the
# registry string to an attribute, such that when reading the file we know
# which reader to use)
#
# When registering read/write capabilities for a user-defined class, we
# recommend adding some random hash to the string identifier to ensure that it
# stays unique for as long as possible. noctiluca uses the hash 198593848. Its
# predecessor tracklib uses the hash 602354027.
reader_writer_registry = {}

def registrystring(obj):
    try:
        return [key for key, (ty, _, _) in reader_writer_registry.items() if ty == type(obj)][0]
    except Exception:
        raise RuntimeError(f"Could not find {str(type(obj))} in reader_writer_registry")

def writer(func):
    def wrapper(obj, name, hdf5_base):
        container = func(obj, name, hdf5_base)
        if container:
            container.attrs[_TYPE_KW] = registrystring(obj)
        return container
    return wrapper

def write(obj, name, hdf5_base):
    try:
        my_reg = registrystring(obj)
    except RuntimeError:
        if np.isscalar(obj):
            hdf5_base.attrs[name] = obj
            return None
        elif isinstance(obj, (h5py.SoftLink, h5py.HardLink, h5py.ExternalLink,
                              h5py.Group, h5py.Dataset,
                              )):
            hdf5_base[name] = obj
            return obj
        else:
            raise

    _, my_write, _ = reader_writer_registry[my_reg]
    return my_write(obj, name, hdf5_base)

def read(hdf5_container):
    try:
        my_reg = hdf5_container.attrs[_TYPE_KW]
    except KeyError:
        if isinstance(hdf5_container, h5py.Group):
            my_reg = 'dict'
        else: # pragma: no cover
            raise RuntimeError(f"Do not understand format of {hdf5_container.name}")

    _, _, my_read = reader_writer_registry[my_reg]
    return my_read(hdf5_container)

###################### Library for built-ins and numpy #########################

def new_group(hdf5_base, name):
    # This can be used to write directly to the root instead of creating new
    # groups
    if name is None:
        return hdf5_base
    return hdf5_base.create_group(name)

@writer
def write_dict(obj, name, hdf5_base):
    group = new_group(hdf5_base, name)
    for key in obj:
        _ = write(obj[key], key, group)
    return group

def read_group_as_dict(group):
    out = {name : read(group[name]) for name in group}
    out.update(group.attrs)

    try:
        del out[_TYPE_KW]
    except KeyError:
        pass

    return out

reader_writer_registry['dict'] = (dict, write_dict, read_group_as_dict)

@writer
def write_ndarray(obj, name, hdf5_base):
    if len(obj.shape) > 0:
        dset = hdf5_base.create_dataset(name, data=obj)
    else:
        hdf5_base.attrs[name] = obj
        dset = None
    return dset

def read_dataset_as_ndarray(dset):
    return np.asarray(dset)

reader_writer_registry['np.ndarray'] = (np.ndarray, write_ndarray, read_dataset_as_ndarray)

@writer
def write_iterable(obj, name, hdf5_base):
    try:
        first_type = type(next(iter(obj)))
    except StopIteration:
        return new_group(hdf5_base, name)

    if (not np.dtype(first_type).kind in ['O', 'U'] # don't try to write numpy arrays for objects or strings
        and all(type(entry) == first_type for entry in obj)
        and name is not None
       ):
        container = hdf5_base.create_dataset(name, data=np.asarray(list(obj)))
    else:
        container = new_group(hdf5_base, name)
        for i, entry in enumerate(obj):
            _ = write(entry, str(i), container)
    return container

def read_iterable(hdf5_container):
    mytype, _, _ = reader_writer_registry[hdf5_container.attrs[_TYPE_KW]]
    if isinstance(hdf5_container, h5py.Dataset):
        return mytype(read_dataset_as_ndarray(hdf5_container))
    else:
        data = read_group_as_dict(hdf5_container)
        return mytype(data[key] for key in sorted(data, key=int))

reader_writer_registry['list'] = (list, write_iterable, read_iterable)
reader_writer_registry['tuple'] = (tuple, write_iterable, read_iterable)
reader_writer_registry['set'] = (set, write_iterable, read_iterable)

@writer
def write_None(obj, name, hdf5_base):
    assert obj is None
    return new_group(hdf5_base, name)

def read_None(group):
    return None

reader_writer_registry['None'] = (type(None), write_None, read_None)

@writer
def write_generic_class(obj, name, hdf5_base):
    # Note that the _HDF5_ORIG_TYPE_ attribute (would be 'dict' for vars())
    # will be overwritten by the @writer decorator
    return write(vars(obj), name, hdf5_base)

def read_generic_class(hdf5_container):
    # This requires that the class can be instantiated without arguments
    cls, _, _ = reader_writer_registry[hdf5_container.attrs[_TYPE_KW]]
    obj = cls()
    obj.__dict__.update(read_group_as_dict(hdf5_container))
    return obj

### Convenience functions when handling HDF5 files ###

def ls(filename, group='/', depth=1):
    """
    List toplevel contents of file (or group within a file)

    Parameters
    ----------
    filename : str or pathlib.Path
        the file to inspect
    group : str
        the group whose contents to list
    depth : int
        how many levels of content to recurse through when encountering
        subgroups

    Returns
    -------
    list of str
        the contents of the specified group, one string per item. Attributes
        are printed with their value and enclosed in braces {}, Datasets are
        surrounded by brackets [], Groups are just given by name.
    """
    def read_group(f, group, remaining_depth, prefix):
        if remaining_depth == 0:
            return []

        contents = []
        for name in f[group]:
            if isinstance(f[group][name], h5py.Dataset):
                contents.append(prefix+'['+name+']')
            else:
                contents.append(prefix+name)
                contents += read_group(f, group+'/'+name, remaining_depth-1, prefix+name+'/')

        for key, value in f[group].attrs.items():
            contents.append(f"{prefix}{{{key} = {str(value)}}}")

        return contents

    group, name = check_group_or_attr(group)
    with h5py.File(str(filename), 'r') as f:
        if name:
            return f[group].attrs[name]
        else:
            return read_group(f, group, depth, '')

def check_group_or_attr(group):
    """
    Check whether we're querying a group or just a single attribute

    The syntax for querying a specific attribute is ``group/{attr}`` (as
    opposed to ``group/subgroup``).

    Parameters
    ----------
    group : str or None
        the identifier to check/dissect

    Returns
    -------
    group, name : str
        if the input conforms to the attribute syntax, it is dissected into
        `!group` and `!name`. Otherwise `!group` is the same as input and
        `!name` is ``None``. Exception: if input is ``None``, the output
        `!group` is ``'/'``.
    """
    if group is None:
        return '/', None

    # regex: should split group/{attr} into group and attr
    #        not match group{attr}, which would be bogus
    #        match /{attr} correctly
    # Note: for "{attr}" (no leading slash), the first group in m will be None
    m = re.fullmatch(r"(?:(.*)/)?{(\w+)}", group)
    name = None
    if m:
        group = m[1] if m[1] and len(m[1]) > 0 else '/'
        name = m[2]

    return group, name



"""
Random assortment of useful auxiliary stuff
"""
import numpy as np
import scipy.interpolate

# This is imported in the util module as *, so don't clutter
"""
exec "norm jjd}O" | let @a="\n'" | exec "g/^def /norm w\"Ayt(:let @a=@a.\"',\\n'\"
" | norm i__all__ = ["ap}kcc]kV?__all__
j>>
"""
__all__ = [
    'distribute_noiselevel',
    'log_derivative',
]

def distribute_noiselevel(noise2, pixelsize):
    """
    Get coordinate-wise localization error from scalar noise level

    Parameters
    ----------
    noise2 : float
        the total noise level. In terms of coordinate-wise errors this is
        ``noise2 = Δx^2 + Δy^2 + Δz^2``
    pixelsize : array-like
        linear extension of pixel/voxel in each direction

    Returns
    -------
    localization_error : (d,) np.array
        localization error distributed to the coordinate directions,
        proportionately to the pixel size, i.e. ``Δx``, ``Δy``, ``Δz`` such
        that ``noise2 = Δx^2 + Δy^2 + Δz^2``.
    """
    voxel = np.asarray(pixelsize)
    noise_in_px = np.sqrt(noise2/np.sum(voxel**2))
    return noise_in_px*voxel

def log_derivative(y, x=None, resampling_density=2):
    """
    Calculate loglog-derivative.

    We resample the given data to log-spaced x.

    Parameters
    ----------
    y : array-like
        the function values whose derivative we are interested in
    x : array-like, optional
        the independent variable for the data in y. Will default to
        ``np.arange(len(y))`` (and thus ignore the first data point).
    resampling_density : float, optional
        how tight to space the log-resampled points. A value of 1 corresponds
        to the spacing between the first two data points, higher values
        decrease spacing.

    Returns
    -------
    x : np.array
        the log-resampled abscissa
    dlog : np.array
        the calculated log-derivative
    """
    if x is None:
        x = np.arange(len(y))

    with np.errstate(divide='ignore'):
        x = np.log(x)
        y = np.log(y)
    ind_valid = np.logical_and(np.isfinite(x), np.isfinite(y))
    x = x[ind_valid]
    y = y[ind_valid]

    dlogx = (x[1] - x[0])/resampling_density
    xnew = np.arange(np.min(x), np.max(x), dlogx)
    ynew = scipy.interpolate.interp1d(x, y)(xnew)

    return np.exp(xnew[:-1] + dlogx/2), np.diff(ynew)/dlogx



import os,sys
try: # prevent any plotting
    del os.environ['DISPLAY']
except KeyError:
    pass
from pathlib import Path

import numpy as np
np.seterr(all='raise') # pay attention to details

import unittest
from unittest.mock import patch

from context import noctiluca as nl
import h5py

"""
exec "norm jjd}O" | let @a="\n'" | exec "g/^class Test/norm w\"Ayt(:let @a=@a.\"',\\n'\"
" | norm i__all__ = ["ap}kcc]kV?__all__
j>>
"""
__all__ = [
    'TestLoad',
    'TestWrite',
    'TestHDF5',
]

class myTestCase(unittest.TestCase):
    def assert_array_equal(self, array1, array2):
        try:
            np.testing.assert_array_equal(array1, array2)
            res = True
        except AssertionError as err: # pragma: no cover
            res = False
            print(err)
        self.assertTrue(res)

class TestLoad(myTestCase):
    def test_evalSPT(self):
        testdata="""
1.0	2.3	1	5	
1.5	2.1	2	5	
2.1	1.5	3	5	
1.9	1.7	5	5	
0.5	9.3	-5	10	
0.4	8.5	-4	10	
1.2	9.1	-6	10	
"""[1:-1] # Cut off '\n' at beginning and end

        path = Path('./testdata')
        path.mkdir(exist_ok=True)
        filename = str(path / "evalSPT.csv")
        with open(filename, mode='w') as f:
            f.write(testdata)

        ds = nl.io.load.evalSPT(filename, tags={'test'})

        ds.makeSelection(selector = lambda traj, _ : len(traj) <= 3)
        self.assert_array_equal(ds[0][:], np.array([[1.2, 9.1], [0.5, 9.3], [0.4, 8.5]]))

        ds.makeSelection(selector = lambda traj, _ : len(traj) > 3)
        self.assert_array_equal(ds[0][:], np.array([[1.0, 2.3], [1.5, 2.1], [2.1, 1.5], [np.nan, np.nan], [1.9, 1.7]]))

    def test_csv(self):
        testdata="""
line,id,noise,x,y,frame,x2,y2,meta_mean,meta_unique,meta_nanmean
1,32,10,0.6,0.8,10,0.5,1.0,1,5,1
2,32,1.5,0.7,0.5,9,0.7,0.8,2,5,nan
3,20,5.3,1.2,1.3,11,-0.8,-1.2,3,6,5
"""[1:-1]

        path = Path('./testdata')
        path.mkdir(exist_ok=True)
        filename = str(path / "twoLocus.csv")
        with open(filename, mode='w') as f:
            f.write(testdata)

        # First two are just dummies
        with self.assertRaises(ValueError): # too many columns specified
            ds = nl.io.load.csv(
                    filename, [None, 'id', 'noise', 'x', 'y', 'z', 't', 'x2', 'y2', 'z2', 'meta_mean', 'meta_unique'],
                    meta_post={'meta_mean' : 'mean', 'meta_unique' : 'unique'}, delimiter=',', skip_header=1)

        with self.assertRaises(AssertionError): # y2 does not have a y to go with it
            ds = nl.io.load.csv(
                    filename, [None, 'id', 'noise', 'x', 't', 'x2', 'y2', 'meta_mean', 'meta_unique'],
                    meta_post={'meta_mean' : 'mean', 'meta_unique' : 'unique'}, delimiter=',', skip_header=1)

        ds = nl.io.load.csv(
                filename, [None, 'id', 'noise', 'x', 'y', 't', 'x2', 'y2', 'meta_mean', 'meta_unique', 'meta_nanmean'],
                meta_post={'meta_mean' : 'mean',
                           'meta_unique' : 'unique',
                           'meta_nanmean' : 'nanmean',
                           }, delimiter=',', skip_header=1)

        ds.makeSelection(selector = lambda traj, _ : len(traj) >= 2)
        self.assert_array_equal(ds[0][:], np.array([[[0.7, 0.5], [0.6, 0.8]], [[0.7, 0.8], [0.5, 1.0]]]))
        self.assert_array_equal(ds[0].meta['noise'], np.array([1.5, 10]))
        self.assertEqual(ds[0].meta['meta_mean'], 1.5)
        self.assertEqual(ds[0].meta['meta_unique'], 5)
        self.assertEqual(ds[0].meta['meta_nanmean'], 1)

        ds.makeSelection(selector = lambda traj, _ : len(traj) < 2)
        self.assert_array_equal(ds[0][:], np.array([[[1.2, 1.3]], [[-0.8, -1.2]]]))
        self.assert_array_equal(ds[0].meta['noise'], np.array([5.3]))
        self.assertEqual(ds[0].meta['meta_mean'], 3)
        self.assertEqual(ds[0].meta['meta_unique'], 6)
        self.assertEqual(ds[0].meta['meta_nanmean'], 5)

        with self.assertRaises(RuntimeError):
            ds = nl.io.load.csv(
                    filename, [None, 'id', 'noise', 'x', 'y', 't', 'x2', 'y2', 'meta_mean', 'meta_unique'],
                    meta_post={'meta_mean' : 'unique', 'meta_unique' : 'unique'}, delimiter=',', skip_header=1)

class TestWrite(myTestCase):
    def setUp(self):
        self.ds = nl.TaggedSet()
        self.ds.add(nl.Trajectory([0, 0.75, 0.5, 0.3, 5.4, 5.5, 5.3, -2.0, 5.4]))
        self.ds.add(nl.Trajectory([1.2, 1.4, np.nan, np.nan, 10.0, 10.2]))

    def test_csv(self):
        path = Path('./testdata')
        path.mkdir(exist_ok=True)
        filename = str(path / 'test_write.csv')
        nl.io.write.csv(self.ds, filename)

        with open(filename, 'r') as f:
            self.assertTrue(f.read() == 'id\tframe\tx\n0\t0\t0.0\n0\t1\t0.75\n0\t2\t0.5\n0\t3\t0.3\n0\t4\t5.4\n0\t5\t5.5\n0\t6\t5.3\n0\t7\t-2.0\n0\t8\t5.4\n1\t0\t1.2\n1\t1\t1.4\n1\t4\t10.0\n1\t5\t10.2\n')

        ds = nl.TaggedSet()
        ds.add(nl.Trajectory(np.arange(20).reshape((2, 5, 2))))
        nl.io.write.csv(ds, filename)

        with open(filename, 'r') as f:
            self.assertTrue(f.read() == 'id\tframe\tx\ty\tx2\ty2\n0\t0\t0\t1\t10\t11\n0\t1\t2\t3\t12\t13\n0\t2\t4\t5\t14\t15\n0\t3\t6\t7\t16\t17\n0\t4\t8\t9\t18\t19\n')

    def test_mat(self):
        path = Path('./testdata')
        path.mkdir(exist_ok=True)
        filename = str(path / 'test_write.mat')
        nl.io.write.mat(self.ds, filename)

class TestHDF5(myTestCase):
    def test_simple(self):
        data = {
            'array' : np.array([1, 2, 3]),
            'noarray' : np.array(3.5),
            'bool' : True,
            'int' : 5,
            'float' : 5.3,
            'complex' : 1+3j,
            'str' : "Hello World",
            'Trajectory' : nl.Trajectory([1, 2, 4, 5], meta_test='moo'),
            'TaggedSet' : nl.TaggedSet(),
            'empty_tuple' : tuple(),
            'None' : None,
        }
        data['TaggedSet'].add(5, ['moo', 'foo', 'bar'])
        data['TaggedSet'].add(8.7)
        data['TaggedSet'].add(nl.Trajectory([1.5, 3.8]), 'traj')

        path = Path('./testdata')
        path.mkdir(exist_ok=True)
        filename = str(path / 'test.hdf5')
        nl.io.write.hdf5(data, filename)

        # Immediately reload
        data_read = nl.io.load.hdf5(filename)
        self.assertEqual(data.keys(), data_read.keys())

        for key in data:
            if key == 'Trajectory':
                traj = data[key]
                traj_read = data_read[key]
                self.assert_array_equal(traj.data, traj_read.data)
                for meta_key in traj.meta:
                    self.assertEqual(traj.meta[meta_key], traj_read.meta[meta_key])
            elif key == 'TaggedSet':
                self.assertListEqual(data[key]._tags, data_read[key]._tags)
                self.assertListEqual(data[key]._selected, data_read[key]._selected)

                data[key].makeSelection(tags='moo')
                data_read[key].makeSelection(tags='moo')
                self.assertEqual(data[key][0], data_read[key][0])
                
                data[key].makeSelection(tags='traj')
                data_read[key].makeSelection(tags='traj')
                self.assert_array_equal(data[key][0].data, data_read[key][0].data)
            elif key == 'array':
                self.assert_array_equal(data[key], data_read[key])
            elif key == 'empty_tuple':
                self.assertEqual(type(data[key]), type(data_read[key]))
                self.assertEqual(len(data_read[key]), 0)
            elif key == 'None':
                self.assertIs(data_read[key], None)
            else:
                self.assertEqual(data[key], data_read[key])

        # Test partial writing
        nl.io.write.hdf5(None, filename, '/None_group/test/')

        # Test partial reading
        self.assertTrue(nl.io.load.hdf5(filename, '/{bool}'))
        self.assertTrue(nl.io.load.hdf5(filename, '/bool'))
        self.assertEqual(nl.io.load.hdf5(filename, '{float}'), data['float'])
        self.assertIsNone(nl.io.load.hdf5(filename, 'None'))
        self.assertEqual(nl.io.load.hdf5(filename, 'empty_tuple/{_HDF5_ORIG_TYPE_}'), 'tuple')

        # Silent overwrite
        nl.io.write.hdf5({}, filename, '/None_group')

    def test_errors(self):
        path = Path('./testdata')
        path.mkdir(exist_ok=True)
        filename = path / 'hdf5_dummy.hdf5'

        class Test:
            pass
        with self.assertRaises(RuntimeError):
            nl.io.write.hdf5(Test(), filename)

        nl.io.write.hdf5(5, filename, group='/test/test')
        res = nl.io.load.hdf5(filename)
        self.assertEqual(res['test']['test'], 5)

    def test_ls(self):
        path = Path('./testdata')
        path.mkdir(exist_ok=True)
        filename = str(path / 'test.hdf5') # this is bad style, relying on a file written in another test...
        ls = nl.io.hdf5.ls(filename)
        self.assertIn('TaggedSet', ls)
        self.assertIn('[array]', ls)
        self.assertIn('{bool = True}', ls)
        self.assertIn('{_HDF5_ORIG_TYPE_ = dict}', ls)

        ls = nl.io.hdf5.ls(filename, '/Trajectory')
        self.assertIn('[data]', ls)
        self.assertIn('meta', ls)

        ls = nl.io.hdf5.ls(filename, depth=2)
        self.assertIn('TaggedSet/[_selected]', ls)
        self.assertIn('Trajectory/meta', ls)
        self.assertIn('empty_tuple/{_HDF5_ORIG_TYPE_ = tuple}', ls)

        self.assertTrue(nl.io.hdf5.ls(filename, '{bool}'))
        self.assertTrue(nl.io.hdf5.ls(filename, '/{bool}'))
        self.assertEqual(nl.io.hdf5.ls(filename, 'empty_tuple/{_HDF5_ORIG_TYPE_}'), 'tuple')

        # Just for completeness
        self.assertTupleEqual(nl.io.hdf5.check_group_or_attr(None), ('/', None))

    def test_write_subTaggetSet(self):
        path = Path('./testdata')
        path.mkdir(exist_ok=True)
        filename = str(path / 'hdf5_dummy.hdf5')

        # Have to make sure that data._data is not converted to numpy array or stored as attributes
        # One way to ensure this is to use dicts, which will be stored as groups
        data = nl.TaggedSet((({'i':i}, 'small' if i < 10 else 'large') for i in range(20)))
        self.assertEqual(len(data), 20)

        nl.io.write.hdf5({}, filename) # empty out the file
        nl.io.write.hdf5(data, filename, 'data_full')
        data.makeSelection(tags='small')

        # A few failing attempts
        with self.assertRaises(ValueError):
            nl.io.write.hdf5_subTaggedSet(data, filename, 'data_small') # forgot refTaggedSet
        with self.assertRaises(ValueError):
            nl.io.write.hdf5_subTaggedSet(data, filename, group='/', refTaggedSet='data_full') # forgot name for new entry

        nl.io.write.hdf5_subTaggedSet(data, filename, 'data_small', refTaggedSet='data_full') # this is how it's done

        read = nl.io.load.hdf5(filename, 'data_small')
        read.makeSelection()
        self.assertEqual(len(read), 10)

        with h5py.File(filename, 'r') as f:
            for i in range(10):
                self.assertEqual(f[f'data_full/_data/{i}'], f[f'data_small/_data/{i}'])
                self.assertNotEqual(f[f'data_full/_tags/{i}'], f[f'data_small/_tags/{i}'])

        # Check that silent overwrite works
        nl.io.write.hdf5_subTaggedSet(data, filename, 'data_small', refTaggedSet='data_full')

        # Test the error cases
        data = nl.TaggedSet(((i, 'small' if i < 10 else 'large') for i in range(20)))
        nl.io.write.hdf5({}, filename) # empty out the file
        nl.io.write.hdf5(data, filename, 'data_full')
        data.makeSelection(tags='small')
        with self.assertRaises(ValueError):
            nl.io.write.hdf5_subTaggedSet(data, filename, 'data_small', refTaggedSet='data_full')

if __name__ == '__main__': # pragma: no cover
    unittest.main(module=__file__[:-3])



"""
Convert user data to noctiluca objects (Trajectory, TaggedSet)

This module provides convenience functions to convert from various formats in
which users might provide their data to the objects that noctiluca and
libraries building on it use. These functions are designed to be used under the
hood in downstream libraries, such that beginning users do not have to go the
extra step of creating Trajectory/TaggedSet objects from their data. In the
long run it is of course recommended to get familiar with these objects, as
they provide some useful additional functionality.

For trajectories, the following inputs are accepted:

+ ``noctiluca.Trajectory``, which is returned unchanged
+ ``pandas.DataFrame``; columns should be named ``'x', 'y', 'z'`` to indicate
  dimension, possibly with a suffix identifying different loci. So for example

  - ``('x', 'y', 'x2', 'y2')`` would indicate a trajectory with two loci in two
    spatial dimensions
  - equivalent column specifications to the above include ``('x_left',
    'x_right', 'y_left', 'y_right')``, ``(x0, y0, x1, y1)``, ``('x_red',
    'y_red', 'x_green', y_green')``
  - the precise regular expression to identify columns containing positional
    data is ``'(x|y|z)(\d*|_.*)'``.
  - for each locus all dimensions should exist, i.e. ``('x0', 'y0', 'x0')``
    would be invalid
  - dimensions should be labelled starting with ``'x'``, i.e. a dataframe
    containing just a ``'y'`` column would be invalid

  Beyond the positional data, the dataframe may contain a column ``'frame'``
  indicating the frame number that each detection belongs to. This column will
  be coerced to integer.
+ numpy array-like; anything that casts well to a numpy array is accepted. This
  case is simply forwarded to the ``noctiluca.Trajectory`` constructor, which
  accepts array-like structures. Possible shapes for array-like inputs:

  - ``(N, T, d)`` for a trajectory with ``N`` loci in ``d`` spatial dimensions
    over ``T`` time steps
  - ``(T, d)`` for a single-locus trajectory
  - ``(T,)`` for a trajectory with a single locus in a single spatial
    dimension.

+ ``pandas.Series``, which is cast to a one dimensional numpy array. In doing
  so, we use the index of the series as frame number, i.e. ``pd.Series(data=[1,
  2, 3], index=[5, 7, 8])`` would be read as a trajectory with 4 frames, where
  the second one is missing.

For datasets (i.e. a collection of trajectories) we accept

+ ``noctiluca.TaggedSet``, which is returned unchanged
+ any iterable of structures that `make_Trajectory` can deal with. Specifically
  this includes a list of numpy arrays (which then can be of different length /
  dimension) or a big numpy array, where the different trajectories are indexed
  along the first dimension.
+ ``pandas.DataFrame`` with column names conforming to the patterns outlined
  above for `make_Trajectory`, plus an additional column ``'particle'``
  indicating which detections belong to which trajectory. The name
  ``'particle'`` for this column was chosen for compatibility with `!trackpy`.
  Note that contrary to `make_Trajectory`, the specification of the ``'frame'``
  column is now mandatory.

More control over the input process (e.g. the assumed column names for pandas
dataframes) can be attained by using the functions in this module explicitly,
in which case additional keyword arguments can be specified. For yet finer
control we recommend directly using the ``noctiluca.Trajectory`` and
``noctiluca.TaggedSet`` classes.
"""
import re

import numpy as np
import pandas as pd

from ..trajectory import Trajectory
from ..taggedset import TaggedSet

def make_Trajectory(inpt, **kwargs):
    """
    Create a ``noctiluca.Trajectory`` from user specified data

    Parameters
    ----------
    inpt : <various>
        see `module docstring <noctiluca.util.userinput>` for description.

    Keyword Arguments
    -----------------
    t_column : str
        name of the column containing frame numbers when giving a
        ``pandas.DataFrame`` as `!inpt`
    pos_columns : (list of) list of str
        names of the columns containing the positional data when giving a
        ``pandas.DataFrame`` as `!inpt`. This should be either a list of
        identifiers for the different dimensions (e.g. ``['x', 'y']``) or a
        list of such lists, if the trajectory has multiple loci (e.g. ``[['x1',
        'y1'], ['x2', 'y2']]``).

    Returns
    -------
    noctiluca.Trajectory
    """
    # Preproc
    if isinstance(inpt, pd.Series):
        inpt = inpt.dropna()
        ind = inpt.index.to_numpy()
        if not np.issubdtype(ind.dtype, np.integer):
            raise ValueError("Could not interpret index of pandas.Series object as integer frame numbers")

        ind -= np.min(ind)
        arr = np.empty(np.max(ind)+1, dtype=float)
        arr[:] = np.nan
        arr[ind] = inpt
        inpt = arr

    # Main processing logic
    if isinstance(inpt, Trajectory):
        return inpt
    elif isinstance(inpt, pd.DataFrame):
        # Find the right entries in the dataframe
        if 't_column' not in kwargs: # str or None, then assume contiguous data
            kwargs['t_column'] = 'frame'
        if 'pos_columns' not in kwargs: # (list of) list of str
            pcols = [col for col in inpt.columns if re.fullmatch(r'(x|y|z)(\d*|_.*)', col)]
            d = max([ord(col[0]) - ord('x') for col in pcols]) + 1
            locus_ids = np.unique([col[1:] for col in pcols])
            N = len(locus_ids)

            pcols_expect = [[chr(ord('x')+dim)+locus_id for dim in range(d)]
                            for locus_id in locus_ids]
            if any(col not in pcols for dcols in pcols_expect for col in dcols) or len(pcols) > N*d:
                raise ValueError(f"Detected positional columns {pcols} are incompatible with expected/inferred names {pcols_expect}")

            kwargs['pos_columns'] = pcols_expect

        # Prepare for assembly
        try:
            frame = inpt[kwargs['t_column']].to_numpy().astype(int)
        except KeyError:
            frame = inpt.index.to_numpy()
            if not np.issubdtype(frame.dtype, np.integer):
                raise ValueError("Could not interpret index of pandas.DataFrame object as integer frame numbers")
        frame = frame - np.min(frame)
        T = np.max(frame) + 1

        pcol = np.array(kwargs['pos_columns'])
        if len(pcol.shape) == 1:
            pcol = np.expand_dims(pcol, 0)
        N, d = pcol.shape

        # Assemble
        data = np.empty((N, T, d), dtype=float)
        data[:] = np.nan
        for n in range(N):
            for dim in range(d):
                data[n, frame, dim] = inpt[pcol[n, dim]]

        return Trajectory(data)
    else: # assume input is castable to numpy array and just expose Trajectory() constructor
        return Trajectory(inpt, **kwargs)
        
def make_TaggedSet(inpt, **kwargs):
    """
    Convert a collection of trajectories to ``noctiluca.TaggedSet``

    Parameters
    ----------
    inpt : <various>
        see `module docstring <noctiluca.util.userinput>` for description.

    Keyword Arguments
    -----------------
    id_column : str
        the name of the column containing trajectory identifiers (when
        specifying a ``pandas.DataFrame``). The data in this column does not
        have to be numerical. Defaults to ``'particle'``.
    **kwargs : <other keyword arguments>
        forwarded to `make_Trajectory`. Note ``'t_column'`` and
        ``'pos_columns'``.
    """
    if isinstance(inpt, TaggedSet):
        return inpt
    elif isinstance(inpt, pd.DataFrame):
        if 'id_column' not in kwargs:
            kwargs['id_column'] = 'particle' # from trackpy
        if 't_column' not in kwargs:
            kwargs['t_column'] = 'frame'

        ids = inpt[kwargs['id_column']].unique()
        out = TaggedSet()
        for id in ids:
            ind = inpt[kwargs['id_column']] == id
            out.add(make_Trajectory(inpt.loc[ind], **kwargs))

        return out
    else: # assume iterable of something that make_Trajectory() can deal with
        return TaggedSet((make_Trajectory(x, **kwargs) for x in inpt), hasTags=False)



import os,sys
try: # prevent any plotting
    del os.environ['DISPLAY']
except KeyError:
    pass

import numpy as np
np.seterr(all='raise') # pay attention to details

import unittest
from unittest.mock import patch

from context import noctiluca
Trajectory = noctiluca.Trajectory
TaggedSet = noctiluca.TaggedSet
del noctiluca

"""
exec "norm jjd}O" | let @a="\n'" | exec "g/^class Test/norm w\"Ayt(:let @a=@a.\"',\\n'\"
" | norm i__all__ = ["ap}kcc]kV?__all__
j>>
"""
__all__ = [
    'Test0Trajectory',
    'Test1Trajectory',
    'Test0TaggedSet',
    'Test1TaggedSet',
]

class myTestCase(unittest.TestCase):
    def assert_array_equal(self, array1, array2):
        try:
            np.testing.assert_array_equal(array1, array2)
            res = True
        except AssertionError as err: # pragma: no cover
            res = False
            print(err)
        self.assertTrue(res)

class Test0Trajectory(myTestCase):
    def test_init(self):
        _ = Trajectory()

        traj = Trajectory(np.zeros((2, 5, 2)), t=[1, 2, 4, 5, 7])
        self.assertEqual(traj.T, 7)
        self.assertTrue(np.all(np.isnan(traj[[2, 5]])))

        traj = Trajectory(0)
        self.assertTupleEqual(traj.data.shape, (1, 1, 1))
        traj = Trajectory([1, 2, 3])
        self.assertTupleEqual(traj.data.shape, (1, 3, 1))
        traj = Trajectory([[1, 2, 3], [4, 5, 6]])
        self.assertTupleEqual(traj.data.shape, (1, 2, 3))
        traj = Trajectory([[[1, 2, 3]], [[1, 2, 3]]])
        self.assertTupleEqual(traj.data.shape, (2, 1, 3))

        with self.assertRaises(ValueError):
            traj = Trajectory(np.zeros((1, 1, 1, 1)))

        with self.assertRaises(ValueError):
            traj = Trajectory([3], t=[[1, 2]])
        with self.assertRaises(ValueError):
            traj = Trajectory([3], t=[1, 2])

        traj = Trajectory(np.zeros((2, 3, 1)), localization_error=849, parity='even', moo=5)
        self.assertEqual(traj.localization_error, 849)
        self.assertEqual(traj.parity, 'even')
        self.assertEqual(traj.meta['moo'], 5)

class Test1Trajectory(myTestCase):
    def setUp(self):
        self.T = 10
        self.Ns = [1, 1, 1, 2, 2, 2]
        self.ds = [1, 2, 3, 1, 2, 3]
        self.trajs = [Trajectory(np.zeros((N, self.T, d)), localization_error=np.ones((d,)), parity='even')
                      for N, d in zip(self.Ns, self.ds)]
        self.trajs[5].localization_error = np.ones((2, 3)) # To check that shape

    def test_count_valid_frames(self):
        traj = Trajectory(np.zeros((2, 5, 2)), t=[1, 2, 4, 5, 7])
        self.assertEqual(traj.count_valid_frames(), 5)

        traj.data[0, 2, :] = 0
        self.assertEqual(traj.count_valid_frames(), 5)

        traj.data[1, 2, :] = 0
        self.assertEqual(traj.count_valid_frames(), 6)

    def test_interface(self):
        for traj, N, d in zip(self.trajs, self.Ns, self.ds):
            # Basic
            self.assertEqual(len(traj), self.T)
            self.assertEqual(traj.N, N)
            self.assertEqual(traj.T, self.T)
            self.assertEqual(traj.T, traj.F) # no missing frames
            self.assertEqual(traj.d, d)

            if N == 1:
                self.assertTupleEqual(traj[3:5].shape, (2, d))
                self.assert_array_equal(traj[2], np.zeros((d,)))
            else:
                self.assertTupleEqual(traj[3:5].shape, (N, 2, d))
                self.assert_array_equal(traj[2], np.zeros((N, d)))

            # Modifiers
            _ = traj.rescale(2)

            traj.meta['MSD'] = {'data' : np.array([5])}
            times2 = traj.rescale(2, keepmeta=['MSD'])
            self.assertIn('MSD', times2.meta)
            self.assert_array_equal(traj.data*2, times2.data)
            plus1 = traj.offset(1, keepmeta=['MSD'])
            self.assert_array_equal(traj.data+1, plus1.data)
            plus1 = traj.offset([1])
            self.assert_array_equal(traj.data+1, plus1.data)
            plus1 = traj.offset([[1]])
            self.assert_array_equal(traj.data+1, plus1.data)

            if N == 2:
                rel = traj.relative(keepmeta=['MSD'])
                self.assertTupleEqual(rel.data.shape, (1, self.T, d))
                self.assertIn('MSD', rel.meta)
                mag = rel.abs(keepmeta=['MSD'])
                self.assertTupleEqual(mag.data.shape, (1, self.T, 1))
                dif = rel.diff(dt=2, keepmeta=['MSD'])
                self.assertTupleEqual(dif.data.shape, (1, self.T-2, d))
                dim = traj.dims([0], keepmeta=['MSD'])
                self.assertTupleEqual(dim.data.shape, (2, self.T, 1))
            elif N == 1:
                mag = traj.abs()
                self.assertNotIn('MSD', mag.meta)
                self.assertTupleEqual(mag.data.shape, (1, self.T, 1))
                with self.assertRaises(ValueError):
                    rel = traj.relative()
                dif = traj.diff(dt=3)
                self.assertTupleEqual(dif.data.shape, (1, self.T-3, d))
                if d >= 2:
                    dim = traj.dims([0, 1])
                    self.assertTupleEqual(dim.data.shape, (1, self.T, 2))

    def test_relative_Nloci(self):
        traj = Trajectory(np.arange(30).reshape((3, 5, 2)))
        dif_sequential = traj.relative()
        self.assert_array_equal(dif_sequential.data, traj.data[1:] - traj.data[:-1])
        dif_ref1 = traj.relative(ref=1)
        self.assert_array_equal(dif_ref1.data, traj.data[[0, 2]] - traj.data[[1]])

        with self.assertRaises(ValueError):
            _ = traj.relative(ref=5)

class Test0TaggedSet(unittest.TestCase):
    def test_init(self):
        ls = TaggedSet()

        ls = TaggedSet(zip([1, 2, 3], [["a", "b"], "a", ["b", "c"]]))
        self.assertListEqual(ls._data, [1, 2, 3])
        self.assertListEqual(ls._tags, [{"a", "b"}, {"a"}, {"b", "c"}])

        ls = TaggedSet([1, 2, 3], hasTags=False)
        
    def test_add(self):
        ls = TaggedSet()
        ls.add(1)
        ls.add(2, 'a')
        ls.add(3, {'b', 'c'})
        
class Test1TaggedSet(unittest.TestCase):
    def setUp(self):
        self.ls = TaggedSet(zip([1, 2, 3], [["a", "b"], "a", ["b", "c"]]))

    def test_len(self):
        self.assertEqual(len(self.ls), 3)
        
    def test_iteration(self):
        for ind, val in enumerate(self.ls):
            self.assertEqual(val, self.ls._data[ind])

        for ind, (val, tags) in enumerate(self.ls(giveTags=True)):
            self.assertEqual(val, self.ls._data[ind])
            self.assertSetEqual(tags, self.ls._tags[ind])

        all_values = {1, 2, 3}
        for val in self.ls(randomize=True):
            all_values.remove(val)
        self.assertSetEqual(all_values, set())

    def test_elementaccess(self):
        self.assertEqual(self.ls[1], 2)

    def test_mergein(self):
        newls = TaggedSet(zip([4, 5, 6], [["d"], [], "e"]))
        self.ls.mergein(newls, additionalTags='new')
        self.assertListEqual(self.ls._data, [1, 2, 3, 4, 5, 6])
        self.assertSetEqual(self.ls.tagset(), {'a', 'b', 'c', 'd', 'e', 'new'})

        self.ls |= newls
        self.assertEqual(len(self.ls), 9)
        
        self.ls.makeSelection(tags='new')
        self.assertSetEqual(set(self.ls), {4, 5, 6})

    def test_makeTagsSet(self):
        self.assertSetEqual(TaggedSet.makeTagsSet("foo"), {"foo"})
        self.assertSetEqual(TaggedSet.makeTagsSet(["foo"]), {"foo"})
        with self.assertRaises(ValueError):
            TaggedSet.makeTagsSet(1)

    def test_selection(self):
        for _ in range(5):
            self.ls.makeSelection(nrand=1, random_seed=6542)
            self.assertEqual(len(self.ls), 1)
        for _ in range(5):
            self.ls.makeSelection(prand=0.5)
            self.assertEqual(len(self.ls), 1)

        self.ls.makeSelection(tags="a")
        self.assertSetEqual({*self.ls}, {1, 2})

        self.ls.makeSelection(tags=["a", "b"], logic=all)
        self.assertSetEqual({*self.ls}, {1})

        self.ls.makeSelection(tags=["a", "b"], logic=any)
        self.assertSetEqual({*self.ls}, {1, 2, 3})

        def sel(datum, tags):
            return datum >= 2
        self.ls.makeSelection(selector=sel)
        self.assertSetEqual({*self.ls}, {2, 3})

        self.ls.refineSelection(tags='c')
        self.assertSetEqual({*self.ls}, {3})

        sel = self.ls.saveSelection()
        self.assertListEqual(sel, [False, False, True])

        self.ls.makeSelection()
        self.assertEqual(len(self.ls), 3)
        self.ls.restoreSelection(sel)
        self.assertEqual(len(self.ls), 1)
        self.assertSetEqual({*self.ls}, {3})

        copied = self.ls.copySelection()
        copied.makeSelection()
        self.assertEqual(len(copied), 1)

    def test_deleteSelection(self):
        self.ls.makeSelection(tags='a')
        self.ls.deleteSelection()
        self.assertEqual(len(self.ls), 1)
        self.assertEqual(self.ls[0], 3)

    def test_addTags(self):
        self.ls.addTags('moo')
        self.assertSetEqual(self.ls.tagset(), {'a', 'b', 'c', 'moo'})

    def test_tagset(self):
        self.assertSetEqual(self.ls.tagset(), {'a', 'b', 'c'})

    def test_apply(self):
        def fun(i):
            return i+1
        self.ls.apply(fun)
        self.assertListEqual(self.ls._data, [2, 3, 4])

    def test_process(self):
        # need a TaggedSet of mutable objects
        mutls = TaggedSet(zip(['hello', 'World', '!'], [["a", "b"], "a", ["b", "c"]]))
        mutls.makeSelection(tags="a")
        newls = mutls.process(lambda word : word+'_moo')

        self.assertListEqual(mutls._data, ['hello', 'World', '!'])
        self.assertListEqual(newls._data, ['hello_moo', 'World_moo'])

    def test_map_unique(self):
        def funTrue(x):
            return True
        def fun2(x):
            return 2*x

        self.assertTrue(self.ls.map_unique(funTrue))
        with self.assertRaises(RuntimeError):
            self.ls.map_unique(fun2)
        self.assertIsNone(TaggedSet().map_unique(funTrue))

if __name__ == '__main__': # pragma: no cover
    unittest.main(module=__file__[:-3])



"""
A small module for running MCMC sampling. It provides the abstract base class
`Sampler` that can be subclassed for concrete sampling schemes.
"""

from copy import deepcopy
from abc import ABC, abstractmethod

from tqdm.auto import tqdm

import numpy as np

class MCMCRun:
    """
    The results of an MCMC run

    Attributes
    ----------
    logLs : np.array(dtype=float)
        the likelihoods from the run
    samples : list of <sample data type>)
        the actual MCMC sample

    Notes
    -----
    The list of samples is not automatically copied upon initialization.
    """
    # Implementation note: `samples` should be a list, such that we can take
    # advantage of mutable objects not being copied.
    def __init__(self, logLs=None, samples=None):
        if logLs is None:
            logLs = np.array([])
        self.logLs = np.asarray(logLs)

        if samples is None:
            samples = []
        self.samples = samples

    def logLs_trunc(self):
        """
        Give only the likelihoods associated with the samples

        `Sampler.run` returns likelihoods starting at the first iteration, but
        samples only after a given burn-in period. This function cuts this
        initial overhang from the likelihood array

        Returns
        -------
        (len(samples),) np.ndarray
        """
        return self.logLs[-len(self.samples):]

    def best_sample_logL(self):
        """
        Give the best sample (and log-likelihood)

        Returns
        -------
        sample : <sample data type>
            the maximum likelihood estimate
        logL : float
            the maximum likelihood value
        """
        logLs = self.logLs_trunc()
        i_best = np.argmax(logLs)
        return self.samples[i_best], logLs[i_best]

    def acceptance_rate(self, criterion='sample_identity', comp_fun=None):
        """
        Calculate fraction of accepted moves

        We can see whether a move was accepted or rejected by checking whether
        the samples are actually different. There are three different ways to
        do so: syntactically correct would be sample comparison (using ``==``).
        However, since samples have a user-defined data type, we do not
        necessarily have the == operator defined. For mutable objects we can
        exploit that they would not be copied for an unaccepted step, i.e. we
        can use identity check (``is``). As a last resort, we can use the
        likelihood as a proxy: it is very unlikeliy that we did a move where
        the likelihood remained exactly the same.

        Alternatively, you can provide a comparison function
        ``comp_fun(sample0, sample1)->bool`` that provides the comparison. This
        overrides the `!criterion` setting.

        Parameters
        ----------
        criterion : {'sample_equality', 'sample_identity', 'likelihood_equality'}
            which method to use to determine whether a step was accepted or
            rejected.
        comp_fun : callable, optional
            should return ``True`` if the two arguments compare equal,
            ``False`` otherwise

        Returns
        -------
        float
            the calculated acceptance rate
        """
        if comp_fun is None:
            if criterion == 'sample_equality':
                def comp_fun(sample0, sample1): return sample0 == sample1
            elif criterion == 'sample_identity':
                def comp_fun(sample0, sample1): return sample0 is sample1

        if comp_fun is not None:
            n_reject = np.count_nonzero([comp_fun(sample0, sample1)
                                         for sample0, sample1 in zip(self.samples[:-1],
                                                                     self.samples[1:])])
        elif criterion == 'likelihood_equality':
            logLs = self.logLs_trunc()
            n_reject = np.sum(logLs[:-1] == logLs[1:])
        else: # pragma: no cover
            raise ValueError("Did not understand inputs")

        return 1 - ( float(n_reject) / (len(self.samples)-1) )

    def evaluate(self, fun):
        """
        Evaluate a function on all samples

        This exploits that (if the sample data type is e.g. a user-defined
        class) many samples will actually be identical and we have to evaluate
        the function significantly fewer than ``len(samples)`` times.

        Parameters
        ----------
        fun : callable of signature ``fun(sample) --> object``
            the function to evaluate. It should expect a single sample as input
            and return something.

        Returns
        -------
        list
            a list of output values, in the order of `samples`.

        Notes
        -----
        This function is supposed to decrease computational cost. It is usually
        quicker however, to use a vectorized function ``fun`` instead.
        """
        last_val = fun(self.samples[0])
        last_sample = self.samples[0]
        out = [last_val]
        for sample in self.samples[1:]:
            if sample is not last_sample:
                last_sample = sample
                last_val = fun(sample)
            out.append(last_val)
        return out

class Sampler(ABC):
    """
    Abstract base class for MCMC sampling
    
    To implement a sampling scheme, subclass this and override

    - `propose_update`
    - `logL`
    - `callback_logging` (optional)
    - `callback_stopping` (optional)
    
    See their documentations in this base class for more details.

    The above methods take an argument `!params` that is supposed to represent
    the parameters of the model. The structure of this representation (e.g. a
    list of parameters, a `!np.ndarray`, or a custom class) is completely up to
    the user. Since all explicit handling of the parameters happens in the
    `propose_update` and `logL` methods, the code in this base class is
    completely independent from the parameterization.

    Attributes
    ----------
    config : dict
        some configuration values. See `configure`.

    Example
    -------

    >>> import numpy as np
    ... import scipy.stats

    >>> class normMCMC(Sampler):
    ...     def __init__(self, stepsize=0.1):
    ...         self.stepsize = stepsize
    ...
    ...     def propose_update(self, current_value):
    ...         proposed_value = current_value + np.random.normal(scale=self.stepsize)
    ...         logp_forward = -0.5*(proposed_value - current_value)**2/self.stepsize**2 - 0.5*np.log(2*np.pi*self.stepsize**2)
    ...         logp_backward = -0.5*(current_value - proposed_value)**2/self.stepsize**2 - 0.5*np.log(2*np.pi*self.stepsize**2)
    ...         return proposed_value, logp_forward, logp_backward
    ...         
    ...     def logL(self, value):
    ...         return -0.5*value**2 - 0.5*np.log(2*np.pi)
    ...     
    ...     def callback_logging(self, current_value, best_value):
    ...         print("Current value: {}".format(current_value))

    >>> mc = normMCMC()
    ... mc.configure(iterations=100, burn_in=50, log_every=10, show_progress=True)
    ... logL, vals = mc.run(1)

    """

    @abstractmethod
    def propose_update(self, params):
        """
        Propose an update step.

        Parameters
        ----------
        params : <user-specified parameter structure>
            the current parameters

        Returns
        -------
        proposed_values : <user-specified parameter structure>
            the proposed new parameters
        logp_forward : float
            (log of the) probability of proposing these values from the current
            ones
        logp_backward : float
            (log of the) probability of proposing the current values from the
            proposed ones

        See also
        --------
        logL, callback_logging
        """
        pass # pragma: no cover
        
    @abstractmethod
    def logL(self, params):
        """
        Calculate log-likelihood of the given parameters.

        Parameters
        ----------
        params : <user-specified parameter structure>
            the parameters to use

        Returns
        -------
        float
            the log-likelihood for the given parameters.

        See also
        --------
        propose_update, callback_logging
        """
        pass # pragma: no cover
    
    def callback_logging(self, current_params, best_params):
        """
        Callback upon logging.

        This function will be called whenever a status line is printed. It can
        thus be used to provide additional information about where the sampler
        is, or how the current values compare to the best ones found so far,
        etc.

        Parameters
        ----------
        current_params : <user-specified parameter structure>
            the current parameter set.
        best_params : <user-specified parameter structure>
            the best (i.e. highest likelihood) parameter set the sampler has
            seen so far.

        See also
        --------
        propose_update, logL
        """
        pass # pragma: no cover

    def callback_stopping(self, myrun):
        """
        Callback to enable early stopping

        This function will be called at well-defined intervals to check whether
        sampling should abort. Judgement is to be made based on the data so
        far, which is handed over as the `!myrun` argument. This is an
        `MCMCRun` object.

        Parameters
        ----------
        myrun : MCMCRun
            the data generated so far

        Returns
        -------
        stop : bool
            should be ``True`` if the sampling is to stop, ``False`` otherwise
        """
        pass # pragma: no cover
    
    def configure(self,
            iterations=1,
            burn_in=0,
            log_every=-1,
            check_stopping_every=-1,
            show_progress=False,
            ):
        """
        Set the configuration of the sampler.

        Note that after calling this function once, you can also access single
        configuration entries via the attribute `config`.

        Parameters
        ----------
        iterations : int
            how many MCMC iterations to run total
        burn_in : int
            how many steps to discard at the beginning
        log_every : int
            print a status line every ... steps. Set to ``-1`` (default) to
            disable status lines.
        check_stopping_every : int
            how often to call the `callback_stopping`. Set to ``-1`` (default)
            to never check.
        show_progress : bool
            whether to show a progress bar using `!tqdm`

        See also
        --------
        run
        """
        self.config = {
                'iterations' : iterations,
                'burn_in' : burn_in,
                'log_every' : log_every,
                'check_stopping_every' : check_stopping_every,
                'show_progress' : show_progress,
                }

    def run(self, initial_values):
        """
        Run the sampling scheme

        Remember to run `configure` before this.
        
        Parameters
        ----------
        initial_values : <user-specified parameter structure>
            the initial values for the sampling scheme
            
        Returns
        -------
        MCMCRun
            the output data of the run. This is essentially a wrapper class for
            an array of likelihoods and the associated samples. Plus some
            convenience functions.

        See also
        --------
        configure

        Notes
        -----
        The returned `MCMCRun` contains likelihoods for all steps starting at
        initialization, while the list of sampled parameters starts only after
        the burn-in period.
        """
        # Input processing
        try:
            config = self.config
        except AttributeError: # pragma: no cover
            raise RuntimeError("Trying to run MCMC sampler before calling configure()")
#         config['_logging'] = config['log_every'] > 0
        
        current_values = initial_values
            
        # Setup
        cur_logL = self.logL(current_values)
        cnt_accept = 0
        cnt_logging = config['log_every']
        cnt_stopping = config['check_stopping_every']
        if cnt_stopping > 0:
            cnt_stopping += config['burn_in']

        max_logL = cur_logL
        best_values = deepcopy(current_values)

        # Have both of these be lists for now, because appending is easier
        # Converted to np.array when handed to MCMCRun
        logLs = []
        params = []
        
        Mrange = range(config['iterations'])
        if config['show_progress']: # pragma: no cover
            Mrange = tqdm(Mrange)

        # Run
        for i in Mrange:
            # Update counters
            cnt_logging -= 1
            cnt_stopping -= 1
                
            # Proposal
            proposed_values, logp_forward, logp_backward = self.propose_update(current_values)
            new_logL = self.logL(proposed_values)
            if np.isfinite(cur_logL):
                with np.errstate(under='ignore', over='ignore'):
                    p_accept = np.exp(new_logL - cur_logL - logp_forward + logp_backward)
            else:
                if cur_logL == -np.inf:
                    # Always move on, even if new_logL might be -inf as well
                    p_accept = 1
                elif cur_logL == np.inf:
                    print("Encountered absorbing state (L = inf) at {}".format(current_values))
                    if i > config['burn_in']: # pragma: no cover
                        print("Aborting here")
                        break
                    else:
                        raise RuntimeError("Absorbing state during burn-in")
                else: # i.e. nan
                    raise RuntimeError("Encountered invalid state (L = NaN) at {}".format(current_values))

            # Acceptance
            if np.random.rand() < p_accept:
                cnt_accept += 1
                current_values = proposed_values
                cur_logL = new_logL
            accept_rate = cnt_accept / (i+1)
            
            # Output
            params.append(current_values)
            logLs.append(cur_logL)
            
            # Keeping track of best parameters
            if cur_logL > max_logL:
                max_logL = deepcopy(cur_logL)
                best_values = deepcopy(current_values)
            
            # Logging
            if cnt_logging == 0 or (i+1 == config['iterations'] and config['log_every'] > 0):
                cnt_logging = config['log_every']
                logstring = "iteration {}: acceptance = {:.0f}%, logL = {}".format(i+1, accept_rate*100, cur_logL)
                print(logstring)
                self.callback_logging(current_values, best_values)

            # Check early stopping
            if cnt_stopping == 0:
                cnt_stopping = config['check_stopping_every']
                if self.callback_stopping(MCMCRun(logLs, params)):
                    break

        # Return
        return MCMCRun(logLs, params[config['burn_in']:])



import os,sys
try: # prevent any plotting
    del os.environ['DISPLAY']
except KeyError:
    pass

import numpy as np
np.seterr(all='raise') # pay attention to details
import pandas as pd

import unittest
from unittest.mock import patch

from context import noctiluca as nl

"""
exec "norm jjd}O" | let @a="\n'" | exec "g/^class Test/norm w\"Ayt(:let @a=@a.\"',\\n'\"
" | norm i__all__ = ["ap}kcc]kV?__all__
j>>
"""
__all__ = [
    'TestClean',
    'TestUtil',
    'TestMCMC',
    'TestMCMCMCMCRun',
    'TestParallel',
]

class myTestCase(unittest.TestCase):
    def assert_array_equal(self, array1, array2):
        try:
            np.testing.assert_array_equal(array1, array2)
            res = True
        except AssertionError as err: # pragma: no cover
            res = False
            print(err)
        self.assertTrue(res)

class TestClean(myTestCase):
    def setUp(self):
        # Split with threshold = 2 => trajectories of lengths [2, 3, 3]
        self.traj = nl.Trajectory([1., 2, 4.1, 4.5, 3, -0.5, -1, -0.7])

        # Split with threshold = 3 => 3 trajectories with lengths [4, 5, 6]
        self.ds = nl.TaggedSet()
        self.ds.add(nl.Trajectory([0, 0.75, 0.5, 0.3, 5.4, 5.5, 5.3, -2.0, 5.4]))
        self.ds.add(nl.Trajectory([1.2, 1.4, np.nan, np.nan, 10.0, 10.2]))

    def test_split_trajectory(self):
        split_trajs = nl.util.clean.split_trajectory_at_big_steps(self.traj, 2)
        self.assert_array_equal(np.sort([len(traj) for traj in split_trajs]), np.array([2, 3, 3]))

    def test_split_dataset(self):
        split_ds = nl.util.clean.split_dataset_at_big_steps(self.ds, 3)
        self.assert_array_equal(np.sort([len(traj) for traj in split_ds]), np.array([4, 5, 6]))

class TestUtil(myTestCase):
    def test_distribute_noiselevel(self):
        self.assert_array_equal(nl.util.distribute_noiselevel(14, [5, 10, 15]),
                                [1, 2, 3])

    def test_log_derivative(self):
        x = [1, 2, 5, np.nan, 30]
        y = np.array(x)**1.438
        xnew, dy = nl.util.log_derivative(y, x, resampling_density=1.5)
        self.assertTrue(np.all(np.abs(dy - 1.438) < 1e-7))

        xnew, dy = nl.util.log_derivative(y)

class TestMCMC(myTestCase):
    myprint = print

    @patch('builtins.print')
    def test_mcmc(self, mock_print):
        # Simply use the example from the docs
        class normMCMC(nl.util.mcmc.Sampler):
            callback_tracker = 0

            def __init__(self, stepsize=0.1):
                self.stepsize = stepsize

            def propose_update(self, current_value):
                proposed_value = current_value + np.random.normal(scale=self.stepsize)
                logp_forward = -0.5*(proposed_value - current_value)**2/self.stepsize**2 - 0.5*np.log(2*np.pi*self.stepsize**2)
                logp_backward = -0.5*(current_value - proposed_value)**2/self.stepsize**2 - 0.5*np.log(2*np.pi*self.stepsize**2)
                return proposed_value, logp_forward, logp_backward
                
            def logL(self, value):
                return -0.5*value**2 - 0.5*np.log(2*np.pi)
            
            def callback_logging(self, current_value, best_value):
                self.callback_tracker += np.abs(current_value) + np.abs(best_value)
                print("test")

            def callback_stopping(self, myrun):
                return len(myrun.logLs) > 7
        
        mc = normMCMC()
        mc.configure(iterations=10,
                     burn_in=5,
                     log_every=2,
                     show_progress=False,)
        myrun = mc.run(1)

        self.assertEqual(mock_print.call_count, 10)
        self.assertEqual(len(myrun.logLs), 10)
        self.assertEqual(len(myrun.samples), 5)
        self.assertGreater(mc.callback_tracker, 0)

        # Stopping
        mc.config['check_stopping_every'] = 2
        myrun = mc.run(1)

        self.assertEqual(len(myrun.logLs), 9)
        self.assertEqual(len(myrun.samples), 4)

        mc.config['check_stopping_every'] = -1

        # infinite likelihoods
        mc.logL = lambda value: -np.inf
        myrun = mc.run(1)
        self.assertEqual(np.sum(np.diff(myrun.samples) != 0), 4)

        mc.logL = lambda value: np.nan
        with self.assertRaises(RuntimeError):
            myrun = mc.run(1)

        mc.logL = lambda value: np.inf
        with self.assertRaises(RuntimeError):
            myrun = mc.run(1)

class TestMCMCMCMCRun(myTestCase):
    def setUp(self):
        sample0 = [0]
        sample1 = [1]
        sample2 = [2]

        self.samples = 5*[sample0] + 2*[sample1] + 3*[sample2]
        self.logLs = np.array(10*[-10] + 5*[-5] + 2*[-2] + 3*[-3])
        self.myrun = nl.util.mcmc.MCMCRun(self.logLs, self.samples)

    def test_init_noparams(self):
        myrun = nl.util.mcmc.MCMCRun()
        self.assertEqual(len(myrun.samples), 0)
        myrun.samples.append(1)
        myrun = nl.util.mcmc.MCMCRun()
        self.assertEqual(len(myrun.samples), 0)

    def test_logLs_trunc(self):
        self.assert_array_equal(self.myrun.logLs_trunc(), self.logLs[10:20])

    def test_best_sample_logL(self):
        best, bestL = self.myrun.best_sample_logL()
        self.assertEqual(bestL, -2)
        self.assertListEqual(best, [1])

    def test_acceptance_rate(self):
        acc = 2/9
        self.assertEqual(self.myrun.acceptance_rate('sample_equality'), acc)
        self.assertEqual(self.myrun.acceptance_rate('sample_identity'), acc)
        self.assertEqual(self.myrun.acceptance_rate('likelihood_equality'), acc)

    def test_evaluate(self):
        tmp = self.myrun.evaluate(lambda sample : 2*sample)
        self.assertIs(tmp[0], tmp[1])
        self.assertTupleEqual(np.array(tmp).shape, (10, 2))

class TestParallel(myTestCase):
    def test_vanilla(self):
        self.assertIs(nl.util.parallel._map, map)
        self.assertIs(nl.util.parallel._umap, map)

        # These tests are a bit of an abuse...
        # This will be used for actual parallelization when testing msdfit
        with nl.util.parallel.Parallelize(dict):
            self.assertIs(nl.util.parallel._map, dict)
            self.assertIs(nl.util.parallel._umap, dict)

        with nl.util.parallel.Parallelize(set, tuple):
            self.assertIs(nl.util.parallel._map, set)
            self.assertIs(nl.util.parallel._umap, tuple)

        self.assertIs(nl.util.parallel._map, map)
        self.assertIs(nl.util.parallel._umap, map)

class TestUserinput(myTestCase):
    def test_make_Trajectory(self):
        make_Trajectory = nl.util.userinput.make_Trajectory

        dat = nl.Trajectory([1, 2, 3])
        self.assertIs(make_Trajectory(dat), dat)

        dat = pd.DataFrame(data=[1, 2, 3], columns=['x'])
        traj = make_Trajectory(dat)
        self.assertEqual(traj.T, 3)
        with self.assertRaises(ValueError):
            make_Trajectory(pd.DataFrame(data=[1, 2, 3], columns=['y']))
        with self.assertRaises(ValueError):
            make_Trajectory(pd.DataFrame(data=[1, 2, 3], columns=['x'], index=['a', 4, 5]))

        dat = pd.DataFrame(data=[[1, 2, 3], [4, 5, 6]], columns=['x', 'y', 'w'])
        traj = make_Trajectory(dat)
        self.assertEqual(traj.d, 2)
        traj = make_Trajectory(dat, pos_columns=['y', 'w'])
        self.assertEqual(traj.d, 2)
        self.assertEqual(traj[0][1], 3)
        traj = make_Trajectory(dat, pos_columns=[['y'], ['x']])
        self.assertEqual(traj.N, 2)

        dat = pd.Series(data=[1, 2, 3], index=[4, 5, 7])
        traj = make_Trajectory(dat)
        self.assertTrue(np.isnan(traj[2][0]))
        with self.assertRaises(ValueError):
            make_Trajectory(pd.Series(data=[1, 2, 3], index=[1, 1.5, 2]))

    def test_make_TaggedSet(self):
        make_TaggedSet = nl.util.userinput.make_TaggedSet

        dat = nl.TaggedSet([1, 2, 3], hasTags=False)
        self.assertIs(make_TaggedSet(dat), dat)

        dat = pd.DataFrame(data=[[1, 2, 3, 1, 1],
                                 [4, 5, 6, 1, 2],
                                 [7, 8, 9, 1, 4],
                                 [9, 8, 7, 2, 5],
                                 [6, 5, 4, 2, 6],
                                ],
                           columns=['x', 'y', 'z', 'particle', 'frame'],
                          )
        ds = make_TaggedSet(dat)
        self.assertEqual(len(ds), 2)
        self.assertEqual(len(ds[0]), 4)

        dat = ([i] for i in range(10))
        ds = make_TaggedSet(dat)
        self.assertEqual(len(ds), 10)

if __name__ == '__main__': # pragma: no cover
    unittest.main(module=__file__[:-3])



"""
Parallelization for noctiluca functions

Use this as context for parallel-aware functions:
>>> import noctiluca as nl
... from multiprocessing import Pool
...
... with Pool(16) as mypool:
...     with nl.util.Parallelize(mypool.imap):
...         # parallel-aware stuff goes here

Note that ``Pool.imap`` is preferred over ``Pool.map``, since it gives an
iterator instead of a list, so is the more literal replacement of the built-in
``map()``. If the code in question is okay with this, you can also use
``Pool.imap_unordered``.

We indicate in docstrings that certain functions are "parallel-aware
(ordered|unordered)" if they use the ``map_like`` or ``map_like_unordered``
arguments, respectively.
"""
_map = map
_umap = map

class Parallelize:
    """
    Context manager taking care of the parallelization

    Parameters
    ----------
    map_like : callable
        should act like ``map()``. Example: ``multiprocessing.Pool.imap``
    map_like_unordered : callable
        should act like ``map()``, but might return the result in a different
        order. Example: ``multiprocessing.Pool.imap_unordered``
    """
    def __init__(self, map_like, map_like_unordered=None):
        self.map = map_like
        if map_like_unordered is None:
            self.umap = map_like
        else:
            self.umap = map_like_unordered

    def __enter__(self):
        global _map, _umap
        _map = self.map
        _umap = self.umap

    def __exit__(self, type, value, traceback):
        global _map, _umap
        _map = map
        _umap = map
        return False # raise anything that might have happened



from . import load
from . import write
from . import hdf5

# Register HDF5 I/O for tracklib classes
# Pack everything into a dummy namespace, so we don't pollute the module
# namespace
# This snippet has to be here because it would result in circular imports if
# placed in either trajectory.py/taggedset.py or io/hdf5.py
class stupid_namespace:
    from ..trajectory import Trajectory
    from ..taggedset import TaggedSet

    for cls in [TaggedSet, Trajectory]:
        hdf5.reader_writer_registry["198593848_noctiluca."+cls.__name__] = (cls, hdf5.write_generic_class, hdf5.read_generic_class)

##### Historically used identifiers (should continue to support these) #####
    # 602354027_tracklib.Trajectory_1N1d, 602354027_tracklib.Trajectory_1N2d, 602354027_tracklib.Trajectory_1N3d
    # 602354027_tracklib.Trajectory_2N1d, 602354027_tracklib.Trajectory_2N2d, 602354027_tracklib.Trajectory_2N3d
    # 602354027_tracklib.Trajectory
    # 602354027_tracklib.TaggedSet
    for N in [1, 2]:
        for d in [1, 2, 3]:
            hdf5.reader_writer_registry[f"602354027_tracklib.Trajectory_{N}N{d}d"] = (Trajectory, hdf5.write_generic_class, hdf5.read_generic_class)

    hdf5.reader_writer_registry[f"602354027_tracklib.Trajectory"] = (Trajectory, hdf5.write_generic_class, hdf5.read_generic_class)
    hdf5.reader_writer_registry[f"602354027_tracklib.TaggedSet"] = (TaggedSet, hdf5.write_generic_class, hdf5.read_generic_class)

del stupid_namespace



import os,sys
try: # prevent any plotting
    del os.environ['DISPLAY']
except KeyError:
    pass

import numpy as np
np.seterr(all='raise') # pay attention to details

import unittest
from unittest.mock import patch

from context import noctiluca as nl

"""
exec "norm jjd}O" | let @a="\n'" | exec "g/^class Test/norm w\"Ayt(:let @a=@a.\"',\\n'\"
" | norm i__all__ = ["ap}kcc]kV?__all__
j>>
"""
__all__ = [
    'TestP2',
]

class myTestCase(unittest.TestCase):
    def assert_array_equal(self, array1, array2):
        try:
            np.testing.assert_array_equal(array1, array2)
            res = True
        except AssertionError as err: # pragma: no cover
            res = False
            print(err)
        self.assertTrue(res)

class TestP2(myTestCase):
    def setUp(self):
        self.traj = nl.Trajectory([1, 2, 3, 4, np.nan, 6])
        raw_data = np.cumsum(np.random.normal(size=(10, 10)), axis=0)
        self.ds = nl.TaggedSet([nl.Trajectory(raw_data[n]) for n in range(len(raw_data))], hasTags=False)

    def test_MSDtraj(self):
        nl.analysis.MSD(self.traj)
        msd = self.traj.meta['MSD']['data']
        self.assert_array_equal(msd, np.array([0, 1, 4, 9, 16, 25]))
        self.assert_array_equal(self.traj.meta['MSD']['N'], np.array([5, 3, 3, 2, 1, 1]))

        del self.traj.meta['MSD']
        msd = nl.analysis.MSD(self.traj, TA=False, recalculate=True)
        self.assert_array_equal(msd, np.array([0, 1, 4, 9, np.nan, 25]))
        self.assert_array_equal(self.traj.meta['MSD']['N'], np.array([1, 1, 1, 1, 0, 1]))

        traj3d = nl.Trajectory(np.arange(30).reshape(-1, 3))
        nl.analysis.MSD(traj3d)
        msd = traj3d.meta['MSD']['data']
        self.assert_array_equal(msd, 27*np.arange(len(traj3d))**2)

    def test_MSDdataset(self):
        msd, N = nl.analysis.MSD(self.ds, giveN=True)
        self.assertEqual(len(msd), 10)
        self.assert_array_equal(N, len(self.ds)*np.linspace(10, 1, 10))

        msd, var = nl.analysis.MSD(self.ds, givevar=True)
        self.assert_array_equal(np.isnan(var), 10*[False])

        self.assert_array_equal(msd, nl.analysis.MSD(self.ds))

    def test_covariances_and_correlations(self):
        acov = nl.analysis.ACov(self.traj)
        self.assert_array_equal(acov, np.array([13.2, 20/3, 35/3, 11, 12, 6]))
        self.assert_array_equal(self.traj.meta['ACov']['data'], np.array([13.2, 20/3, 35/3, 11, 12, 6]))

        acorr = nl.analysis.ACorr(self.traj)
        self.assert_array_equal(acorr, np.array([13.2, 20/3, 35/3, 11, 12, 6])/13.2)
        self.assert_array_equal(self.traj.meta['ACorr']['data'], np.array([13.2, 20/3, 35/3, 11, 12, 6])/13.2)

        vacov = nl.analysis.VACov(self.traj)
        self.assert_array_equal(vacov, np.array([1, 1, 1, np.nan, np.nan]))
        self.assert_array_equal(self.traj.meta['VACov']['data'], np.array([1, 1, 1, np.nan, np.nan]))

        vacorr = nl.analysis.VACorr(self.traj)
        self.assert_array_equal(vacov, np.array([1, 1, 1, np.nan, np.nan]))
        self.assert_array_equal(self.traj.meta['VACorr']['data'], np.array([1, 1, 1, np.nan, np.nan]))

    def test_new_p2(self):
        def AD(xm, xn):
            return np.sum(np.abs(xm-xn), axis=-1)

        MAD = nl.analysis.p2.P2traj(self.traj, function=AD, writeto=None)['data']
        with self.assertRaises(KeyError):
            _ = self.traj.meta['P2']
        nl.analysis.p2.P2(self.traj, function=AD)

        self.assert_array_equal(MAD, np.array([0, 1, 2, 3, 4, 5]))
        self.assert_array_equal(MAD, self.traj.meta['P2']['data'])

        _ = nl.analysis.p2.P2dataset(self.ds, function=AD)

if __name__ == '__main__': # pragma: no cover
    unittest.main(module=__file__[:-3])



from copy import deepcopy
import random

class TaggedSet():
    """
    A set with tags for each object.

    This class can be used as an iterator in constructs such as ``for datum in
    myset: ...``. The subset to be iterated over can be adjusted using
    `makeSelection`. If you also need the tags for each datum, use
    ``myset(giveTags=True)``.

    For processing data (in the sense of "applying a function to all of them")
    it is often useful to simply use the built-in `!map`. If the function in
    question has to overwrite the data, you can use `apply`. Note that
    in-place modification can be achieved with `!map()`.

    Parameters
    ----------
    iterable : iterable of data or (datum, tags) pairs, optional
        most useful things to use here: an actual list, or a generator.
        Remember to set `!hasTags` accordingly.
    hasTags : bool, optional
        whether the iterable in the first argument gives data only or (datum,
        tags) pairs.

    Notes
    -----
    For a TaggedSet ``myset``, the following operations are defined:

    ``len(myset)``
        return number of data in current selection
    ``myset(giveTags)``
        return a generator yielding single data or (datum, tags) pairs,
        depending on the value of `!giveTags` (False by default).
    ``iter(myset)``
        shortcut for ``myset(giveTags=False)``. This enables the construction
        ``for datum in myset: ...``
    ``myset[i]``
        element access within the current selection. You can use indices from
        ``0`` to ``len(myset)-1``.
    ``myset |= otherset``
        add the data in `!otherset` into `!myset`. This is a shortcut for
        ``myset.mergein(otherset)``.
    """

    ### Utility ###

    @staticmethod
    def makeTagsSet(tags):
        """
        Input reformatting, mostly internal use.

        Make sure that the frequently used `!tags` argument is a set of strings.

        Parameters
        ----------
        tags : str, list of str, set of str, or None

        Returns
        -------
        set of str
        """
        if tags is None:
            return set()
        elif isinstance(tags, str):
            tags = {tags}
        elif isinstance(tags, list):
            tags = set(tags)
        elif not isinstance(tags, set):
            raise ValueError("Did not understand type of 'tags' ({})".format(str(type(tags))))

        return tags

    ### Setting up ###

    def __init__(self, iterable=[], hasTags=True):
        self._data = []
        self._tags = []
        self._selected = []

        if hasTags:
            for datum, tags in iterable:
                self.add(datum, tags)
        else:
            for datum in iterable:
                self.add(datum)

    def add(self, datum, tags=None):
        """
        Append the given datum with the given tags.

        Parameters
        ----------
        datum : object
            the datum to append
        tags : str, list of str, or set of str, optional
            the tags to attach to this new datum

        Notes
        -----
        The newly added datum will be part of the current selection.
        """
        tags = self.makeTagsSet(tags)

        self._data.append(datum)
        self._tags.append(tags)
        self._selected.append(True)

    ### Basic usage ###

    def __iter__(self):
        return self()

    def __call__(self, giveTags=False, randomize=False):
        """
        Iterate through the current selection of the set.

        Parameters
        ----------
        giveTags : bool, optional
            whether to yield only data or (datum, tags) pairs.
        randomize : bool, optional
            set to ``True`` to randomize the order in which data are yielded.

        Yields
        ------
        datum
            the data in the current selection
        set of str, optional
            the tags associated with the datum

        See also
        --------
        makeSelection
        """
        indices = range(len(self._data))
        if randomize:
            indices = random.sample(indices, k=len(indices))

        for i in indices:
            if self._selected[i]:
                if giveTags:
                    yield (self._data[i], self._tags[i])
                else:
                    yield self._data[i]

    def __len__(self):
        return sum(self._selected)

    def __getitem__(self, ind):
        for i, datum in enumerate(self):
            if i >= ind:
                return datum

    ### Selection system ###

    def makeSelection(self, **kwargs):
        """
        Mark a subset of the current set as 'current selection'. For most
        purposes, the set will behave as if it contained only these data.

        There are multiple ways to select data. Which one is used depends on
        the kwargs given. With increasing preference, these methods are

        - select randomly: use the kwargs `!nrand` or `!prand`
        - select by tag: use the kwargs `!tags` and `!logic`
        - select with a user-specified function: use kwarg `!selector`

        Call this without arguments to reset the selection to the whole
        dataset.

        Parameters
        ----------
        nrand : int
            number of trajectories to select at random
        prand : float, in [0, 1]
            fraction of trajectories to select at random
        random_seed : int, str, or bytes, optional
            seed for random selection. Note that this can be a string.
        tags : str, list of str, or set of str
            the tags to select. How these go together will be determined by
            `!logic`.
        logic : callable with signature bool = logic(<list of bool>)
            the logic for handling multiple tags. Set this to (the built-in)
            `!all` to select the data being tagged with all the given tags, or
            to `!any` (the default) to select the data having any of the given
            tags.
        selector : callable with signature bool = selector(datum, tags)
            should expect the datum and a set of tags as input and return True
            if the datum is to be selected, False otherwise.

        Other Parameters
        ----------------
        refining : bool, optional
            set to True to apply the current selection scheme only to those
            data that are already part of the selection. This can be used to
            successively refine a selection by using different selection
            methods (e.g. first by tag, then by some other criterion).

        See also
        --------
        refineSelection, saveSelection, restoreSelection
        """
        assert len(self._data) == len(self._tags) == len(self._selected)
        if not 'refining' in kwargs.keys():
            kwargs['refining'] = False

        # Define selector function according to given arguments
        if 'selector' in kwargs.keys():
            selector = kwargs['selector']
        elif 'tags' in kwargs.keys():
            kwargs['tags'] = TaggedSet.makeTagsSet(kwargs['tags'])
            if not 'logic' in kwargs.keys():
                kwargs['logic'] = any
            def selector(datum, tags):
                return kwargs['logic']([tag in tags for tag in kwargs['tags']])
        elif 'nrand' in kwargs.keys() or \
             'prand' in kwargs.keys():
            curlen = sum(self._selected) if kwargs['refining'] else len(self._data)
            try:
                nrand = kwargs['nrand']
            except KeyError:
                nrand = int(kwargs['prand']*curlen)
            if 'random_seed' in kwargs.keys():
                random.seed(kwargs['random_seed'])
            toselect = random.sample(range(curlen), nrand)
            def selector(datum, tags):
                # Note: have to hack a bit to get a static variable
                try:
                    return selector.cnt in toselect
                except AttributeError:
                    selector.cnt = 0
                    return selector.cnt in toselect
                finally:
                    selector.cnt += 1
        else:
            def selector(datum, tags):
                return True

        for i, (datum, tags, selected) in enumerate(zip(self._data, self._tags, self._selected)):
            if selected or not kwargs['refining']:
                self._selected[i] = bool(selector(datum, tags))

    def refineSelection(self, *args, **kwargs):
        """
        A wrapper for ``makeSelection(..., refining=True)``.

        See also
        --------
        makeSelection
        """
        kwargs['refining'] = True
        self.makeSelection(*args, **kwargs)

    def saveSelection(self):
        """
        Return a copy of the current selection for reference.

        This is useful when planning to undo subsequent selections without
        having to redo the whole selection from scratch.

        Returns
        -------
        list of bool
            the current selection.

        See also
        --------
        restoreSelection, makeSelection
        """
        return deepcopy(self._selected)

    def restoreSelection(self, selection):
        """
        Restore a previously saved selection.

        Parameters
        ----------
        selection : list of bool
            the selection that was saved using `saveSelection`. Will be copied.

        See also
        --------
        saveSelection, makeSelection
        """
        self._selected = deepcopy(selection)

    def copySelection(self):
        """
        Generate a new (copied) set from the current selection.

        Returns
        -------
        The new set

        See also
        --------
        makeSelection
        """
        def gen():
            for traj, tags in self(giveTags=True):
                yield(deepcopy(traj), deepcopy(tags))

        return TaggedSet(gen())

    def deleteSelection(self):
        """
        Remove the selected entries from the set

        See also
        --------
        makeSelection, copySelection

        Notes
        -----
        Since everything in the current selection is deleted, the selection
        will be reset.
        """
        self._data = [datum for datum, selected in zip(self._data, self._selected) if not selected]
        self._tags = [tagset for tagset, selected in zip(self._tags, self._selected) if not selected]
        self._selected = len(self._data)*[True]

    ### Managing multiple TaggedSets ###

    def mergein(self, other, additionalTags=None):
        """
        Add the contents of the TaggedSet `!other` to the caller.

        Parameters
        ----------
        other : TaggedSet
            the set whose data to add
        additionalTags : str, list of str, or set of str, optional
            additional tag(s) to add to all of the new data.

        Notes
        -----
        This can also be invoked as ``self |= other``. In that case no
        additional tags can be added.
        """
        if not issubclass(type(other), TaggedSet): # pragma: no cover
            raise TypeError("Can only merge a TaggedSet")

        additionalTags = TaggedSet.makeTagsSet(additionalTags)
        newTags = [tags | additionalTags for tags in other._tags]

        self._data += other._data
        self._tags += newTags
        self._selected += other._selected

    def __ior__(self, other):
        self.mergein(other)
        return self

    ### Global tag management ###

    def addTags(self, tags):
        """
        Add new tag(s) to all data in the current selection

        Parameters
        ----------
        tags : str, list of str, or set of str
            the tag(s) to add
        """
        tags = TaggedSet.makeTagsSet(tags)

        for _, curtags in self(giveTags=True):
            curtags |= tags

    def tagset(self):
        """
        Return the set of all tags in the current selection.

        Returns
        -------
        set of str
            set of all tags in the current selection
        """
        tagset = set()
        for _, tags in self(giveTags=True):
            tagset |= tags

        return tagset


    ### Processing data in the set ###

    def apply(self, fun):
        """
        Apply the given function to all data.

        Parameters
        ----------
        fun : callable of signature ``datum = fun(datum)``

        See also
        --------
        process, builtins.map

        Notes
        -----
        This function works in-place. If you need the original data to remain
        unchanged, use ``process``.
        """
        # Have to explicitly run through the data array, because the entries
        # might be reassigned.
        for i, (datum, selected) in enumerate(zip(self._data, self._selected)):
            if selected:
                self._data[i] = fun(datum)

    def process(self, fun):
        """
        Generate a new `TaggedSet` with processed data.

        Same as `apply`, except that a new set with the processed data is
        returned, while the original one remains unchanged.

        Parameters
        ----------
        fun : callable of signature ``datum = fun(datum)``

        Returns
        -------
        TaggedSet
            a new set containing the processed data

        See also
        --------
        apply, builtins.map

        Notes
        -----
        The new set will contain only the processed data, i.e. data that are
        not in the current selection will not be copied.
        """
        def gen(origin):
            for datum, tags in origin(giveTags=True):
                yield (fun(deepcopy(datum)), deepcopy(tags))

        return TaggedSet(gen(self))

    def map_unique(self, fun):
        """
        Apply a function to all data and check that the result is unique.

        Parameters
        ----------
        fun : callable with signature ``ret = fun(datum)``
            the function to apply. The return value can be any object for which
            ``ret1 == ret2`` is defined.

        Returns
        -------
        ret : object
            if `!fun` returns the same value on all data, that value.

        Raises
        ------
        RuntimeError
            if the return value of `!fun` is not the same on all data.

        See also
        --------
        builtins.map

        Notes
        -----
        Returns ``None`` if ``len(self) == 0`` (i.e. the caller / its current
        selection does not contain any data)
        """
        it = map(fun, self)
        try:
            first = next(it)
        except StopIteration:
            return None
        if all(first == rest for rest in it):
            return first
        else:
            raise RuntimeError("TaggedSet.map_unique() called on data that do not give uniform values")

#     def map(self, fun):
#         """
#         Apply fun to all data and return a list of the results.
# 
#         Parameters
#         ----------
#         fun : callable that takes a datum as argument
# 
#         Returns
#         -------
#         A list of fun(datum) for all data in the list
# 
#         Notes
#         -----
#          - If fun does not provide a return value, the output will be a list of
#            None's. Use _ = ... to ignore this output if necessary.
#          - This can be used to manipulate the data in-place; if fun does not
#            manipulate the datum in place but has signature datum = fun(datum),
#            then use apply() instead.
#          - python's built-in map() also works very well with TaggedSets.
#         """
#         return [fun(traj) for traj in self]



from copy import deepcopy

import numpy as np

class Trajectory:
    """
     + kwargs are written to meta
     + (N, T, d), (T, d), (T,) array for data
     + generic meta entries got converted to attributes
     + localization error is (d,), (N, d), (N, T, d)
    """

    ### Set up ###

    def __init__(self, data=None, t=None, *, localization_error=None, parity=None, **kwargs):
        if data is None:
            self.data = None
        else:
            data = np.array(data) # Note that this also copies the data
            if len(data.shape) > 3:
                raise ValueError("Array of shape {} cannot be interpreted as trajectory".format(str(data.shape)))
            elif len(data.shape) == 2:
                data = np.expand_dims(data, 0)
            elif len(data.shape) == 1:
                data = np.expand_dims(data, (0, 2))
            elif len(data.shape) == 0:
                data = np.expand_dims(data, (0, 1, 2))

            if t is None:
                self.data = data
            else:
                t = np.asarray(t)
                if len(t.shape) > 1:
                    raise ValueError(f"Argument t should be 1d array, but has shape {t.shape}")
                if len(t) != data.shape[1]:
                    raise ValueError(f"len(t) = {len(t)} is not equal to number of timepoints in data ({data.shape[1]})")

                if np.issubdtype(t.dtype, np.integer):
                    t = t - np.min(t)
                    data_patched = np.empty((data.shape[0], np.max(t)+1, data.shape[2]), dtype=float)
                    data_patched[:] = np.nan
                    data_patched[:, t, :] = data
                    self.data = data_patched
                else: # pragma: no cover
                    # This might be implemented at some point, the idea being that
                    # from given floating point timestamps one could identify a
                    # characteristic / median lag time and use that to rectify the
                    # timestamps into integer frames.
                    # While being useful for some applications, this does not
                    # produce a faithful representation of the input data, so maybe
                    # should not run by default without any comment / warning.
                    # Beyond the point above, handing a floating point array to the
                    # `t` argument also has a high likelihood of being
                    # unintentional, so throwing an error might be the best
                    # behavior.
                    raise NotImplementedError("Trajectory() does currently not support float time stamps")

        self.localization_error = localization_error
        self.parity = parity
        self.meta = kwargs

    ### Basic properties ###

    @property
    def N(self):
        """ Number of loci """
        return self.data.shape[0]

    @property
    def T(self):
        """ Length in frames """
        return self.data.shape[1]

    @property
    def F(self):
        """ Number of valid frames """
        return np.sum(~np.any(np.isnan(self.data), axis=(0, 2)))

    @property
    def d(self):
        """ Number of dimensions """
        return self.data.shape[2]

    def __len__(self):
        return self.T

    def count_valid_frames(self):
        """
        Return the number of frames that have data.

        We regard a frame as unusable as soon as any data is missing. So this
        function counts the number of frames in the trajectory where none of
        the data is ``np.nan``.

        Returns
        -------
        int

        See also
        --------
        F
        """
        return self.F

    def __getitem__(self, key):
        """
        Element-access

        Parameters
        ----------
        key : index or slice into the `T` dimension
        
        Returns
        -------
        np.ndarray
            the corresponding part of the trajectory.

        Notes
        -----
        The `N` dimension will be squeezed, i.e. for single locus trajectories
        that first dimension will be removed. The `T` dimension of the output
        follows the numpy conventions (i.e. depends on the format of `!key`),
        while `d` is guaranteed to be present.
        """
        ret = self.data[:, key, :]
        # T squeezing is already done by numpy element-access
        try:
            ret = np.squeeze(ret, axis=0)
        except ValueError:
            pass
        return ret

    ### Modifiers ###

    def abs(self, order=None, keepmeta=None):
        """
        Modifier: 2-norm (or other norm, see `!order`)

        Parameters
        ----------
        order : float or None, optional
            order of the norm; see ``numpy.linalg.norm``'s `!ord` parameter.
            Defaults to 2-norm (i.e. Euclidean norm)
        keepmeta : list of str or None, optional
            which entries from the `meta` dict to copy to the new trajectory

        Returns
        -------
        Trajectory

        Notes
        -----
        For multi-locus trajectories, this will take the norm of each locus
        individually. To get a `Trajectory` of relative distance, use
        ``Trajectory.relative().abs()``.

        See also
        --------
        diff, dims, relative
        """
        traj = Trajectory(np.linalg.norm(self.data, ord=order, axis=2, keepdims=True))

        if keepmeta is not None:
            for key in keepmeta:
                traj.meta[key] = deepcopy(self.meta[key])

        if self.localization_error is not None:
            traj.localization_error = np.linalg.norm(self.localization_error, ord=order, axis=-1, keepdims=True)
        if self.parity is not None:
            traj.parity == 'even'

        return traj

    def diff(self, dt=1, keepmeta=None):
        """
        Modifier: displacements

        Calculate the displacements over `!dt` frames.

        Parameters
        ----------
        dt : integer
            the time lag to use for displacement calculation
            default: 1, i.e. frame to frame displacements
        keepmeta : list of str or None, optional
            which entries from the `meta` dict to copy to the new trajectory

        Returns
        -------
        Trajectory

        See also
        --------
        abs, dims, relative
        """
        traj =  Trajectory(self.data[:, dt:, :] - self.data[:, :-dt, :])

        if keepmeta is not None:
            for key in keepmeta:
                traj.meta[key] = deepcopy(self.meta[key])

        if self.localization_error is not None:
            traj.localization_error = self.localization_error * np.sqrt(2)
        if self.parity is not None:
            traj.parity == 'even' if self.parity == 'odd' else 'odd'
        
        return traj

    def dims(self, key, keepmeta=None):
        """
        Modifier: select dimensions

        Parameters
        ----------
        key : list of int, or slice
            which dimensions to use. Attention: this cannot be a single `!int`. To
            get the ``i``-th spatial component, use ``traj.dims([i])``.
        keepmeta : list of str or None, optional
            which entries from the `meta` dict to copy to the new trajectory

        Returns
        -------
        Trajectory

        See also
        --------
        abs, diff, relative
        """
        traj = Trajectory(self.data[:, :, key])

        if keepmeta is not None:
            for mkey in keepmeta:
                traj.meta[mkey] = deepcopy(self.meta[mkey])

        if self.localization_error is not None:
            traj.localization_error = self.localization_error.take(key, axis=-1)
        traj.parity == self.parity

        return traj

    def rescale(self, factor, keepmeta=None):
        """
        Modifier: rescale all data by a constant factor

        Parameters
        ----------
        factor : float
        
        Returns
        -------
        Trajectory
        keepmeta : list of str or None, optional
            which entries from the `meta` dict to copy to the new trajectory

        See also
        --------
        abs, offset
        """
        traj = Trajectory(self.data * factor)

        if keepmeta is not None:
            for key in keepmeta:
                traj.meta[key] = deepcopy(self.meta[key])

        if self.localization_error is not None:
            traj.localization_error = self.localization_error * factor
        traj.parity = self.parity

        return traj

    def offset(self, off, keepmeta=None):
        """
        Modifier: shift the trajectory by some offset

        Parameters
        ----------
        off : float or array of shape ``(d,)``, ``(N, d)``, or ``(N, T, d)``
            the offset to add
        keepmeta : list of str or None, optional
            which entries from the `meta` dict to copy to the new trajectory

        Returns
        -------
        Trajectory

        See also
        --------
        abs, rescale
        """
        off = np.array(off)
        ls = len(off.shape)
        if ls == 0: # happens if input is scalar
            off = np.expand_dims(off, (0, 1, 2))
        elif ls == 1:
            off = np.expand_dims(off, (0, 1))
        elif ls == 2:
            off = np.expand_dims(off, 1)

        traj = Trajectory(self.data + off)

        if keepmeta is not None:
            for key in keepmeta:
                traj.meta[key] = deepcopy(self.meta[key])

        traj.localization_error = self.localization_error
        traj.parity = self.parity

        return traj

    def relative(self, ref=None, keepmeta=None):
        """
        Modifier: Return `Trajectory` of pairwise distances

        Parameters
        ----------
        ref : int or None, optional
            which locus to use as reference. If ``None`` (the default) return
            sequential differences (i.e. ``2-1, 3-2, 4-3, ...``). Otherwise,
            returned distances are ``1-ref, 2-ref, ...``
        keepmeta : list of str or None, optional
            which entries from the `meta` dict to copy to the new trajectory

        Returns
        -------
        Trajectory

        Raises
        ------
        ValueError
            when called on trajectories with only a single locus

        See also
        --------
        abs, diff, dims
        """
        if self.N == 1:
            raise ValueError("relative() does not apply to single locus trajectories")

        loc = self.localization_error.copy() if self.localization_error is not None else np.array([0.])
        locN = len(loc.shape) > 1
        if not locN:
            loc = np.array(self.N*[loc])
        loc = loc**2

        if ref is None:
            traj = Trajectory(np.diff(self.data, axis=0))
            loc = np.sqrt(loc[:-1] + loc[1:])
        else:
            if not ref < self.N:
                raise ValueError(f"Cannot use locus {ref} as reference for {self.N}-locus trajectory")
            ind = [i for i in range(self.N) if i != ref]
            traj = Trajectory(self.data[ind] - self.data[[ref]])
            loc = np.sqrt(loc[ind] + loc[[ref]])

        if keepmeta is not None:
            for key in keepmeta:
                traj.meta[key] = deepcopy(self.meta[key])

        if self.localization_error is None:
            traj.localization_error = None
        elif not locN:
            traj.localization_error = loc[0]
        else:
            traj.localization_error = loc

        traj.parity = self.parity

        return traj
